from typing import TYPE_CHECKING, Any, Dict, Tuple

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

# Standard library imports
import numpy as np
import scipy as sp

# Third-party library imports
import plotly.graph_objects as go
import matplotlib.colors as mcolors
import matplotlib.cm as cm
from skimage.measure import marching_cubes

# Local application imports
from comfit.tool import (
    tool_complete_field,
    tool_colormap, 
    tool_set_plot_axis_properties_plotly,
    tool_plotly_colorbar,
    tool_plotly_define_2D_plot_ax,
    tool_plotly_define_3D_plot_ax
)


def plot_complex_field_plotly(
        self: 'BaseSystem',
        complex_field: np.ndarray,
        **kwargs: Any
        ) -> Tuple[go.Figure, Dict]:
    """Plot a complex field using Plotly.

    Creates a visualization of a complex field using various plotting methods
    depending on the dimensionality of the field.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    complex_field : np.ndarray
        The complex field to plot
    \*\*kwargs : Any
        Additional keyword arguments to customize the plot

    Returns
    -------
    go.Figure
        The Plotly figure containing the plot
    """

    kwargs['colormap'] = kwargs.get('colormap', 'angle') # Override the default colormap with 'angle'
    complex_field, fig, ax, kwargs = self.plot_prepare(complex_field, field_type = 'complex', **kwargs)

    # Calculate the magnitude and phase of the complex field
    rho = np.abs(complex_field)
    theta = np.angle(complex_field)

    plt_colormap_object = tool_colormap(kwargs['colormap'], plot_lib='matplotlib')

    ###############################################################
    ###################### DIMENSION: 1 ###########################
    ###############################################################

    if self.dim == 1:

        ax = tool_plotly_define_2D_plot_ax(fig, ax)

        # Keyword arguments particular to the 1D case
        grid = kwargs.get('grid', False)

        vlim = kwargs.get('vlim', None)
        if vlim is not None:
            kwargs['ylim'] = vlim

        # Color in the graph based on the argument of the complex field
        blend_factor=0.3 # The degree to which the color is blended with white
        
        color = (1-blend_factor)*np.array(plt_colormap_object((theta[0] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
        color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')

        fig.add_trace(go.Scatter(x=[self.xmin/self.a0,(self.xmin+self.dx/2)/self.a0], 
                        y=[rho[0],(rho[0]+rho[1])/2],
                        mode='lines',
                        line=dict(color='rgba(0,0,0,0)'),
                        fill='tozeroy',
                        showlegend=False,
                        hoverinfo='skip',
                        fillcolor=color_str,
                        xaxis=ax['xN'],
                        yaxis=ax['yN']))

        for i in range(1,self.xRes-1):
            color = (1-blend_factor)*np.array(plt_colormap_object((theta[i] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
            color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')
            fig.add_trace(go.Scatter(x=[(self.x[i]-self.dx/2)/self.a0,self.x[i]/self.a0], 
                        y=[(rho[i]+rho[i-1])/2,rho[i]],
                        mode='lines',
                        line=dict(color='rgba(0,0,0,0)'),
                        fill='tozeroy',
                        showlegend=False,
                        hoverinfo='skip',
                        fillcolor=color_str, 
                        xaxis=ax['xN'],
                        yaxis=ax['yN']))

            color = (1-blend_factor)*np.array(plt_colormap_object((theta[i] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
            color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')
            fig.add_trace(go.Scatter(x=[self.x[i]/self.a0,(self.x[i]+self.dx/2)/self.a0], 
                        y=[rho[i],(rho[i]+rho[i+1])/2],
                        mode='lines',
                        line=dict(color='rgba(0,0,0,0)'),
                        fill='tozeroy',
                        showlegend=False,
                        hoverinfo='skip',
                        fillcolor=color_str,
                        xaxis=ax['xN'],
                        yaxis=ax['yN']))

        color = (1-blend_factor)*np.array(plt_colormap_object((theta[-1] + np.pi) / (2 * np.pi)))+blend_factor*np.array([1,1,1,1])
        color_str = ('rgb('+str(int(color[0]*255))+','+str(int(color[1]*255))+','+str(int(color[2]*255))+')')
        fig.add_trace(go.Scatter(x=[(self.xmax-1.5*self.dx)/self.a0,(self.xmax-self.dx)/self.a0], 
                        y=[(rho[-1]+rho[-2])/2,rho[-1]],
                        mode='lines',
                        line=dict(color='rgba(0,0,0,0)'),
                        fill='tozeroy',
                        showlegend=False,
                        hoverinfo='skip',
                        fillcolor=color_str,
                        xaxis=ax['xN'],
                        yaxis=ax['yN']))


        fig.add_trace(go.Scatter(
            x=self.x/self.a0,
            y=rho,
            mode='lines',
            showlegend=False,
            customdata=np.stack((theta/np.pi, rho), axis=-1),
            hovertemplate='x: %{x:.2f} a₀<br>θ: %{customdata[0]:.2f} π<br>ρ: %{customdata[1]:.2e}',
            name='',
            line=dict(color='black'),
            xaxis=ax['xN'],
            yaxis=ax['yN']
        ))

        
    ###############################################################
    ###################### DIMENSION: 2 ###########################
    ###############################################################

    elif self.dim == 2:
        
        # Keyword arguments particular to the 2D case
        plot_method = kwargs.get('plot_method', 'phase_angle')

        # Create a meshgrid
        X, Y = np.meshgrid(self.x, self.y, indexing='ij')

        if plot_method == '3Dsurface':
        
            print("\033[91mWarning: 3D surface plot not yet implemented for Plotly.\033[0m")
            pass

        elif plot_method == 'phase_angle':
            
            ax = tool_plotly_define_2D_plot_ax(fig, ax)

            # Keyword arguments particular to the phase angle plot
            grid = kwargs.get('grid', False)

            rho_normalized = rho / np.max(rho)
            
            norm = mcolors.Normalize(vmin=-np.pi, vmax=np.pi)

            # Create an RGBA image array
            rgba_image = cm.ScalarMappable(norm=norm, cmap=plt_colormap_object).to_rgba(theta)

            # Apply the spatial opacity
            for i in range(rgba_image.shape[0]):
                for j in range(rgba_image.shape[1]):
                    for k in range(3):
                        rgba_image[i, j, k] = 1-(1-rgba_image[i, j, k])*rho_normalized[i, j]  # Set alpha channel

            rgba_image = np.transpose(rgba_image, (1, 0, 2))

            # Convert RGBA image to a format Plotly can understand
            image_data = (rgba_image * 255).astype(np.uint8)


            if fig == None:
                fig = go.Figure()

            trace = go.Image(z=image_data, 
                            dx=self.dx/self.a0, 
                            dy=self.dy/self.a0, 
                            x0=self.xmin/self.a0, 
                            y0=self.ymin/self.a0,
                            hovertemplate='x: %{x:.2f} a₀<br>y: %{y:.2f} a₀<br>θ: %{customdata[0]:.2f} π<br>ρ: %{customdata[1]:.2e}',
                            customdata=np.stack((np.transpose(theta/np.pi), np.transpose(rho)), axis=-1),
                            name='',
                            xaxis=ax['xN'],
                            yaxis=ax['yN']
                            )

            fig.add_trace(trace)


    ###############################################################
    ###################### DIMENSION: 3 ###########################
    ###############################################################
    elif self.dim == 3:

        ax = tool_plotly_define_3D_plot_ax(fig, ax)

        # Keyword arguments particular to the 3D case
        plot_method = kwargs.get('plot_method', 'phase_blob')
        
        X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')
        rho_normalized = rho / np.max(rho)

        if plot_method == 'phase_angle':
        
            print('\033[91mWarning:\033[0m Plotly not yet implemented for 3D phase angle plots.')
            pass
        
        elif plot_method == 'phase_blob':
            
            phase_blob_threshold = kwargs.get('phase_blob_threshold', 0.5)

            if np.nanmin(rho_normalized)<phase_blob_threshold<np.nanmax(rho_normalized):
                verts, faces, _, _ = marching_cubes(rho_normalized, phase_blob_threshold)

                # Calculate the centroids of each triangle
                centroids = np.mean(verts[faces], axis=1)

                # Assuming theta is defined on the same grid as rho
                x, y, z = np.mgrid[0:rho_normalized.shape[0], 0:rho_normalized.shape[1], 0:rho_normalized.shape[2]]

                # Flatten the grid for interpolation
                points = np.c_[x.ravel(), y.ravel(), z.ravel()]
                theta_values = theta.ravel()

                # Interpolate theta at the vertices positions
                reals = np.real(complex_field)
                imags = np.imag(complex_field)

                interpolation_method = kwargs.get('interpolation_method', 'linear')
                print("Interpolating points with method ", interpolation_method, ".")
                print("If this process is slow, consider passing 'interpolation_method='nearest' with the plot_complex_field function.")
                print("That will speed up the process, but the plot may look less smooth.")
                reals_faces = sp.interpolate.griddata(points, reals.ravel(), centroids, method='nearest')
                imags_faces = sp.interpolate.griddata(points, imags.ravel(), centroids, method='nearest')
                print("Interpolation done.")
                
                theta_faces = np.arctan2(imags_faces, reals_faces)

                # Normalize theta values for color mapping
                theta_faces_normalized = (theta_faces + np.pi) / (2*np.pi)

                # Map normalized theta values to colors
                colors = plt_colormap_object(theta_faces_normalized)

                # Convert colors to 'rgba()' format required by Plotly
                colors = ['rgba({},{},{},{})'.format(*c) for c in (255*colors).astype(int)]

                # Create the mesh object
                x_new = (verts[:, 0] * self.dx + self.xmin) / self.a0
                y_new = (verts[:, 1] * self.dy + self.ymin) / self.a0
                z_new = (verts[:, 2] * self.dz + self.zmin) / self.a0

                mesh = go.Mesh3d(
                    x=x_new, 
                    y=y_new, 
                    z=z_new,
                    i=faces[:, 0], 
                    j=faces[:, 1], 
                    k=faces[:, 2],
                    facecolor=colors,  # Set color for each face
                    showscale=True,
                    scene=ax['sceneN']
                )

                fig.add_trace(mesh)

    
    if kwargs['colorbar'] and not(ax['colorbar']) and not(kwargs['field_is_nan']):
        ax['colormap_object'] = kwargs['colormap_object']
        fig.add_trace(tool_plotly_colorbar(ax, type='angle'))
        ax['colorbar'] = True

    kwargs['fig'] = fig
    kwargs['ax'] = ax
    tool_set_plot_axis_properties_plotly(self, **kwargs)

    return fig, ax


# General packages 
import numpy as np
import plotly.graph_objects as go

def tool_plotly_find_next_xN(
        fig : go.Figure
        ) -> str:
    """Find the next available x-axis identifier in a Plotly figure.

    Parameters
    ----------
    fig : plotly.graph_objects.Figure
        The Plotly figure to analyze.

    Returns
    -------
    str
        The next available x-axis identifier. Returns 'x' if no x-axes are used,
        or 'xN' where N is one more than the current highest axis number.
    """
    used_xaxis = set()
    
    # Check all traces in the figure
    for trace in fig.data:
        if hasattr(trace, 'xaxis'):
            used_xaxis.add(trace.xaxis)
    
    return 'x' if not used_xaxis else f'x{len(used_xaxis)+1}'

if __name__ == "__main__":
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=[1, 2, 3], y=[4, 5, 6], xaxis='x1'))
    fig.add_trace(go.Scatter(x=[1, 2, 3], y=[4, 5, 6], xaxis='x2'))
    fig.add_trace(go.Scatter(x=[1, 2, 3], y=[4, 5, 6], xaxis='x3'))
    fig.add_trace(go.Scatter(x=[1, 2, 3], y=[4, 5, 6], xaxis='x4'))
    fig.add_trace(go.Scatter(x=[1, 2, 3], y=[4, 5, 6], xaxis='x5'))
    fig.add_trace(go.Scatter(x=[1, 2, 3], y=[4, 5, 6], xaxis='x6'))
    print(tool_plotly_find_next_xN(fig)) 
    print('Expected: x7')

    fig = go.Figure()
    print(tool_plotly_find_next_xN(fig)) 
    print('Expected: x')


from typing import List, Tuple, Union, TYPE_CHECKING, Any
if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

import numpy as np
import plotly.graph_objects as go

def tool_set_plot_axis_properties_plotly(
        self : 'BaseSystem', 
        **kwargs : Any
        ) -> None:
    """Set the properties of the axis for a plot.

    Parameters
    ----------
    kwargs : dict
        Keyword arguments for the axis properties.

    Returns
    -------
    None
        Function updates the layout and scene dictionaries with the axis properties.    
    """
    # Create dictionaries to store the layout updates
    layout_updates = {}
    
    # For 3D plots, create dictionaries to store the scene updates
    scene_updates = {}
    xaxis_updates = {}
    yaxis_updates = {}
    zaxis_updates = {}

    ##### FIGURE #####
    if 'fig' not in kwargs:
        raise ValueError("'fig' parameter is required")
    if 'ax' not in kwargs:
        raise ValueError("'ax' parameter is required")
    
    fig = kwargs['fig']
    ax = kwargs['ax']

    ##### SIZE #####
    size = kwargs.get('size', None)

    if size is not None:
        layout_updates['width'] = size[0]
        layout_updates['height'] = size[1]

    ##### PLOT NATURE #####
    row = ax['row']
    col = ax['col']

    plot_is_3D = kwargs.get('plot_is_3D', False)

    ##### GRID #####
    grid = kwargs.get('grid', True)
    if ax['plot_dimension'] == 2:
        layout_updates[f'{ax["xaxisN"]}_showgrid'] = grid
        layout_updates[f'{ax["yaxisN"]}_showgrid'] = grid

    ##### AXIS EQUAL #####
    axis_equal = kwargs.get('axis_equal', False if self.dim==1 else True)

    if axis_equal:
        if ax['plot_dimension'] == 2:
            layout_updates[ax['yaxisN']] = dict(scaleanchor=ax['xN'], scaleratio=1)
        elif ax['plot_dimension'] == 3:
            scene_updates['aspectmode'] = 'cube'
    

    ##### AXIS LIMITS #####
    # xlim is specified as a list
    xlim = [self.xmin/self.a0, (self.xmax-self.dx)/self.a0]
    if 'xmin' in kwargs:
        xlim[0] = kwargs['xmin'] / self.a0
    if 'xmax' in kwargs:
        xlim[1] = kwargs['xmax'] / self.a0
    if 'xlim' in kwargs:
        xlim = np.array(kwargs['xlim']) / self.a0

    # ylim is specified as a list if dim>1 else as None
    ylim = [self.ymin/self.a0, (self.ymax-self.dy)/self.a0] if self.dim > 1 else None
    if 'ymin' in kwargs:
        ylim[0] = kwargs['ymin'] / self.a0 if self.dim > 1 else kwargs['ymin']

    if 'ymax' in kwargs:
        ylim[1] = kwargs['ymax'] / self.a0 if self.dim > 1 else kwargs['ymax']

    if 'ylim' in kwargs:
        ylim = np.array(kwargs['ylim'])/self.a0 if self.dim > 1 else kwargs['ylim']
    elif 'vlim' in kwargs:
        if self.dim == 1:
            ylim = kwargs['vlim']

    # zlim is specified as a list if dim>2 else as None
    zlim = [self.zmin/self.a0, (self.zmax-self.dz)/self.a0] if self.dim > 2 else None
    if 'zmin' in kwargs:
        zlim[0] = kwargs['zmin'] / self.a0 if self.dim > 2 else kwargs['zmin']
    if 'zmax' in kwargs:
        zlim[1] = kwargs['zmax'] / self.a0 if self.dim > 2 else kwargs['zmax']
    if 'zlim' in kwargs:
        zlim = np.array(kwargs['zlim'])/self.a0 if self.dim > 2 else kwargs['zlim']

    if ax['plot_dimension'] == 2:
        layout_updates['xaxis_range'] = xlim
        layout_updates['yaxis_range'] = ylim
    
    elif ax['plot_dimension'] == 3:
        xaxis_updates['range'] = xlim
        yaxis_updates['range'] = ylim
        zaxis_updates['range'] = zlim
        

    ##### AXIS LABELS #####
    xlabel = kwargs.get('xlabel', 'x/a₀')
    ylabel = kwargs.get('ylabel', 'y/a₀' if self.dim > 1 else None)
    zlabel = kwargs.get('zlabel', 'z/a₀' if self.dim > 2 else None)

    if ax['plot_dimension'] == 2:
        layout_updates[f"{ax['xaxisN']}_title"] = xlabel
        if ylabel is not None:
            layout_updates[f"{ax['yaxisN']}_title"] = ylabel

    elif ax['plot_dimension'] == 3:
        xaxis_updates['title'] = xlabel
        yaxis_updates['title'] = ylabel
        zaxis_updates['title'] = zlabel

        
    ##### TICKS #####
    xticks = kwargs.get('xticks', None)  
    xticklabels = kwargs.get('xticklabels', None)

    yticks = kwargs.get('yticks', None)
    yticklabels = kwargs.get('yticklabels', None)

    zticks = kwargs.get('zticks', None)
    zticklabels = kwargs.get('zticklabels', None)

    
    if xticks is not None:
        xaxis_updates['tickvals'] = xticks
    if xticklabels is not None:
        xaxis_updates['ticktext'] = xticklabels

    if yticks is not None:
        yaxis_updates['tickvals'] = yticks
    if yticklabels is not None:
        yaxis_updates['ticktext'] = yticklabels

    if zticks is not None:
        zaxis_updates['tickvals'] = zticks
    if zticklabels is not None:
        zaxis_updates['ticktext'] = zticklabels

    ##### UPDATE SCENE #####
    if ax['plot_dimension'] == 3:
        if xaxis_updates:
            scene_updates['xaxis'] = xaxis_updates

        if yaxis_updates:
            scene_updates['yaxis'] = yaxis_updates

        if zaxis_updates:
            scene_updates['zaxis'] = zaxis_updates

    else:
        if xaxis_updates:
            layout_updates['xaxis'] = xaxis_updates

        if yaxis_updates:
            layout_updates['yaxis'] = yaxis_updates

    ##### UPDATE LAYOUT #####
    # if ax is None:

    ##### ADJUST SUBPLOT POSITION #####
    padding = 0.15

    x_domain_start = (ax['col']-1)/ax['ncols']+padding*(ax['col']-1)/ax['ncols']
    x_domain_end = ax['col']/ax['ncols']-padding*(1-ax['col']/ax['ncols'])
    y_domain_start = 1-ax['row']/ax['nrows']+padding*(ax['nrows']-ax['row'])/ax['nrows']
    y_domain_end = 1-(ax['row']-1)/ax['nrows']-padding*(ax['row']-1)/ax['nrows']

    if ax['plot_dimension'] == 2:
        fig.update_layout({ax['xaxisN']: dict( domain=[x_domain_start, 
                                                        x_domain_end],
                                                        anchor=ax['yN'])})

        fig.update_layout({ax['yaxisN']: dict(domain=[y_domain_start, 
                                                      y_domain_end],
                                                            anchor=ax['xN'])})

        fig.update_layout(layout_updates)

    elif ax['plot_dimension'] == 3:
        scene_updates['domain'] = {'x': [x_domain_start, x_domain_end],
                                   'y': [y_domain_start, y_domain_end]}
                                   
        fig.update_layout({ax['sceneN']: scene_updates})

    ##### TITLE #####
    title = kwargs.get('title', None)
    
    if title is not None:
        fig.add_annotation(x=(x_domain_start), 
        y=y_domain_end+0.05, 
        xref='paper',
        yref='paper',
        text=title, 
        showarrow=False, 
        align='left')
    

    # else:
    #     dummy_fig = go.Figure()
    #     dummy_fig.update_layout(layout_updates)
    #     if plot_is_3D:
    #         dummy_fig.update_layout(scene=scene_updates)

    #     row = int(ax[0,0])
    #     nrows = int(ax[0,1])
    #     col = int(ax[1,0])
    #     ncols = int(ax[1,1])

    #     if not plot_is_3D:
    #         fig.update_xaxes(dummy_fig.layout.xaxis, row = row, col = col)
    #         fig.update_yaxes(dummy_fig.layout.yaxis, row = row, col = col)
        
    #     else:
    #         fig.update_scenes(dummy_fig.layout.scene, rows = [row], cols = [col])


from typing import TYPE_CHECKING, Any, Tuple, Union, List

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

# Standard library imports
from typing import Optional, Union

# Third-party imports
import numpy as np
import plotly.graph_objects as go

# Local imports
from comfit.tool.tool_set_plot_axis_properties_plotly import tool_set_plot_axis_properties_plotly
from comfit.tool import (
    tool_complete_field,
    tool_add_spacing_2D,
    tool_add_spacing_3D,
    tool_plotly_define_2D_plot_ax,
    tool_plotly_define_3D_plot_ax,
    tool_plotly_colorbar
)

def plot_vector_field_plotly(
        self: 'BaseSystem',
        vector_field: Union[np.ndarray, List],
        **kwargs: Any
        ) -> Tuple[go.Figure, dict]:
    """Plot the vector field using Plotly.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    vector_field : np.ndarray
        Array containing the vector field components. Shape depends on dimensionality.
    \*\*kwargs : Any
        spacing : int, optional. The spacing for the quiver plot. Defaults to max(self.xRes//20, 1).
        Additional keyword arguments for plot customization, see https://comfitlib.com/Plotting/

    Returns
    -------
    Tuple[go.Figure, dict]
        The figure containing the plot and the axes dictionary containing plot properties.

    Raises
    ------
    Exception
        If an invalid field is provided to the plot_vector_field function.
    """

    spacing = kwargs.get('spacing', max(self.xRes//20,1))

    vector_field, fig, ax, kwargs = self.plot_prepare(vector_field, field_type = 'vector', **kwargs)

    ###############################################################
    ########### DIMENSION: 1 - VECTOR-DIMENSION: 1 ################
    ###############################################################
        
    if self.dim == 1 and vector_field.shape == (1,self.xRes):

        kwargs['colorbar'] = False
        ax = tool_plotly_define_2D_plot_ax(fig, ax)
        
        X, Y = np.meshgrid(self.x, np.array([0]), indexing='ij')
        
        U = np.zeros(X.shape)
        V = np.zeros(X.shape)

        V[:,0] = vector_field[0]

        X,Y,U,V = tool_add_spacing_2D(X,Y,U,V,spacing)
        
        fig, ax = self.plot_field(vector_field[0], **kwargs)

        
    ###############################################################
    ########### DIMENSION: 1 - VECTOR-DIMENSION: 2 ################
    ###############################################################

    elif self.dim == 1 and vector_field.shape == (2,self.xRes):

        ax = tool_plotly_define_3D_plot_ax(fig, ax)
            
        X, Y, Z = np.meshgrid(self.x, np.array([0]), np.array([0]), indexing='ij')

        U = np.zeros(X.shape)
        V = np.zeros(X.shape)
        W = np.zeros(X.shape)

        V[:,0,0] = vector_field[0]
        W[:,0,0] = vector_field[1]

        ax['vmin'] = kwargs.get('vmin', 0)
        ax['vmax'] = kwargs.get('vmax', np.max(np.sqrt(vector_field[0]**2 + vector_field[1]**2)))

        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        if not kwargs['field_is_nan']:
            fig.add_trace(go.Cone(x=X.flatten()/self.a0, 
                                    y=Y.flatten()/self.a0, 
                                    z=Z.flatten()/self.a0, 
                                    u=U.flatten(), 
                                    v=V.flatten(), 
                                    w=W.flatten(), 
                                    colorscale=kwargs['colormap_object'], 
                                    sizemode='scaled', 
                                    sizeref=1, 
                                    showscale=False,
                                    scene = ax['sceneN'],
                                    hovertemplate='x: %{x:.1f} a₀ <br>' +
                                                'ux: %{u:.1f} <br>' +
                                                'uy: %{v:.1f} <br>' +
                                                'uz: %{w:.1f} <br>' +
                                                '|u|: %{customdata:.1f}<extra></extra>',
                                                customdata=np.sqrt(U**2 + V**2 + W**2).flatten()))



        # kwargs['ylim'] = [np.min(vector_field[0]), np.max(vector_field[0])]
        # delta_y = kwargs['ylim'][1] - kwargs['ylim'][0]

        # kwargs['zlim'] = [np.min(vector_field[1]), np.max(vector_field[1])]
        # delta_z = kwargs['zlim'][1] - kwargs['zlim'][0]

        # if delta_y < 0.15*delta_z:
        #     kwargs['ylim'] = [kwargs['ylim'][0] - 0.15*delta_z, kwargs['ylim'][1] + 0.15*delta_z]
        
        # if delta_z < 0.15*delta_y:
        #     kwargs['zlim'] = [kwargs['zlim'][0] - 0.15*delta_y, kwargs['zlim'][1] + 0.15*delta_y]

    ###############################################################
    ########### DIMENSION: 1 - VECTOR-DIMENSION: 3 ################
    ###############################################################

    elif self.dim == 1 and vector_field.shape == (3,self.xRes):
        

        ax = tool_plotly_define_3D_plot_ax(fig, ax)
        
        X, Y, Z = np.meshgrid(self.x, np.array([0]), np.array([0]), indexing='ij')

        U = np.zeros(X.shape)
        V = np.zeros(X.shape)
        W = np.zeros(X.shape)

        # Add the vector field
        U[:,0,0] = vector_field[0]
        V[:,0,0] = vector_field[1]
        W[:,0,0] = vector_field[2]

        # Add spacing
        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2 + V ** 2 + W ** 2))

        ax['vmin'] = kwargs.get('vmin', 0)
        ax['vmax'] = kwargs.get('vmax', np.max(max_vector))

        # Normalizing
        U = U/max_vector
        V = V/max_vector
        W = W/max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*self.size_x/15/self.a0)
        vy_scale = kwargs.get('vy_scale', 1)
        vz_scale = kwargs.get('vz_scale', 1)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V
        W = vz_scale*W

        if not kwargs['field_is_nan']:
            fig.add_trace(go.Cone(x=X.flatten()/self.a0, 
                            y=Y.flatten()/self.a0, 
                            z=Z.flatten()/self.a0, 
                            u=U.flatten(), 
                            v=V.flatten(), 
                            w=W.flatten(), 
                            colorscale=kwargs['colormap_object'], 
                            sizemode='scaled', 
                            sizeref=1, 
                            showscale=False, 
                            scene = ax['sceneN'],
                            hovertemplate='<b>x:</b> %{x:.1f} a₀ <br>' +
                                        '<b>ux:</b> %{u:.1f} <br>' +
                                        '<b>uy:</b> %{v:.1f} <br>' +
                                        '<b>uz:</b> %{w:.1f} <br>' +
                                        '<b>|u|:</b> %{customdata:.1f}<extra></extra>',
                                        customdata=np.sqrt(U**2 + V**2 + W**2).flatten()))

        # kwargs['axis_equal'] = False
        # kwargs['ylim'] = [-1,1]
        # kwargs['zlim'] = [-1,1]

    ###############################################################
    ########### DIMENSION: 2 - VECTOR-DIMENSION: 1 ################
    ###############################################################

    elif self.dim == 2 and vector_field.shape == (1,self.xRes,self.yRes):
  
        ax = tool_plotly_define_2D_plot_ax(fig, ax)

        X, Y = np.meshgrid(self.x, self.y, indexing='ij')

        U = vector_field[0]
        V = np.zeros(X.shape)

        X,Y,U,V = tool_add_spacing_2D(X,Y,U,V,spacing)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2))
        
        ax['vmin'] = kwargs.get('vmin', 0)
        ax['vmax'] = kwargs.get('vmax', np.max(max_vector))

        U = U / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*self.dx/self.a0)

        # Scaling
        U = vx_scale*U

        
        fig = kwargs.get('fig', go.Figure())

        u = U.flatten()
        v = V.flatten()
        
        magnitude = np.sqrt(u**2 + v**2)
        magnitude_normalized = magnitude/max(np.max(magnitude),1e-12)

        angle = np.arctan2(v, u)
        direction = np.array([np.cos(angle), np.sin(angle)]).T
        
        fig.add_trace(go.Scatter(
        x=X.flatten()/self.a0,
        y=Y.flatten()/self.a0,
        xaxis=ax['xN'],
        yaxis=ax['yN'],
        mode='markers',
        showlegend=False,
        marker=dict(symbol='arrow', 
            angle=90-angle.flatten()*180/np.pi, 
            size=2*spacing*magnitude_normalized.flatten(), 
            sizemode='diameter',
            color=magnitude.flatten(), 
            colorscale=kwargs['colormap_object'], 
            showscale=False,
            line=dict(color='black')
            ),
            hovertemplate='<b>x:</b> %{x:.1f}a0<br>' +
                        '<b>y:</b> %{y:.1f}a0<br>' +
                        '<b>ux:</b> %{customdata[0]:.2f}<br>' +
                        '<b>uy:</b> %{customdata[1]:.2f}<br>' +
                        '<b>|u|:</b> %{customdata[2]:.2f}<extra></extra>',
            customdata=np.stack((u.flatten(), v.flatten(),magnitude.flatten() ), axis=-1)  # Adding ux, uy and u as customdata
        )
        )

    ###############################################################
    ########### DIMENSION: 2 - VECTOR-DIMENSION: 2 ################
    ###############################################################

    elif self.dim == 2 and vector_field.shape == (2,self.xRes,self.yRes):

        ax = tool_plotly_define_2D_plot_ax(fig, ax)

        X, Y = np.meshgrid(self.x, self.y, indexing='ij')

        X, Y, U, V = tool_add_spacing_2D(X,Y,vector_field[0],vector_field[1],spacing)

        u = U.flatten()
        v = V.flatten()
        
        magnitude = np.sqrt(u**2 + v**2)

        ax['vmin'] = kwargs.get('vmin', 0)
        ax['vmax'] = kwargs.get('vmax', np.max(magnitude))

        magnitude_max = max(np.max(magnitude),1e-12)
        magnitude_normalized = magnitude/magnitude_max

        angle = np.arctan2(v, u)
        direction = np.array([np.cos(angle), np.sin(angle)]).T
        
        if not kwargs['field_is_nan']:
            fig.add_trace(go.Scatter(
            x=X.flatten()/self.a0,
            y=Y.flatten()/self.a0,
            mode='markers',
            xaxis=ax['xN'],
            yaxis=ax['yN'],
            showlegend=False,
            marker=dict(symbol='arrow', 
                angle=90-angle.flatten()*180/np.pi, 
                size=4*spacing*magnitude_normalized.flatten(), 
                sizemode='diameter',
                color=magnitude.flatten(), 
                colorscale=kwargs['colormap_object'], 
                showscale=False,
                cmin=0,
                cmax=magnitude_max,
                line=dict(color='black')
                ),
                hovertemplate='<b>x:</b> %{x:.2f}a0<br>' +
                    '<b>y:</b> %{y:.2f}a0<br>' +
                    '<b>ux:</b> %{customdata[0]:.2e}<br>' +  
                    '<b>uy:</b> %{customdata[1]:.2e}<br>' +
                    '<b>|u|:</b> %{customdata[2]:.2e}<extra></extra>',
                customdata=np.stack((u.flatten(), v.flatten(), magnitude.flatten()), axis=-1)  # Adding ux, uy and u as customdata
            )
            )


    ###############################################################
    ########### DIMENSION: 2 - VECTOR-DIMENSION: 3 ################
    ###############################################################

    elif self.dim == 2 and vector_field.shape == (3,self.xRes,self.yRes):

        ax = tool_plotly_define_3D_plot_ax(fig, ax)
        
        X, Y, Z = np.meshgrid(self.x, self.y, np.array([0]), indexing='ij')
        U = np.zeros(X.shape)
        V = np.zeros(X.shape)
        W = np.zeros(X.shape)
        
        U[:,:,0] = vector_field[0]
        V[:,:,0] = vector_field[1]
        W[:,:,0] = vector_field[2]

        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        max_vector = np.max(np.sqrt(U ** 2 + V ** 2 + W ** 2))

        ax['vmin'] = kwargs.get('vmin', 0)
        ax['vmax'] = kwargs.get('vmax', np.max(max_vector))

        # Normalizing
        U = U / max_vector
        V = V / max_vector
        W = W / max_vector
        
        # Scale factors
        vx_scale = kwargs.get('vx_scale', 2*spacing*self.size_x/max_vector/self.a0)
        vy_scale = kwargs.get('vy_scale', 2*spacing*self.size_y/max_vector/self.a0)
        vz_scale = kwargs.get('vz_scale', spacing/self.a0)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V
        W = vz_scale*W

        if not kwargs['field_is_nan']:
            fig.add_trace(go.Cone(x=X.flatten()/self.a0, 
                                y=Y.flatten()/self.a0, 
                                z=Z.flatten()/self.a0, 
                                u=U.flatten(), 
                                v=V.flatten(), 
                                w=W.flatten(), 
                                colorscale=kwargs['colormap_object'], 
                                sizemode='scaled', 
                                sizeref=1, 
                                showscale=False,
                                scene=ax['sceneN'],
                                hovertemplate='<b>x:</b> %{x:.1f} a₀ <br>' +
                                                '<b>y:</b> %{y:.1f} a₀ <br>' +
                                                '<b>z:</b> %{z:.1f} a₀ <br>' +
                                                '<b>ux:</b> %{u:.1f} <br>' +
                                                '<b>uy:</b> %{v:.1f} <br>' +
                                                '<b>uz:</b> %{w:.1f} <br>' +
                                                '<b>|u|:</b> %{customdata:.1f}<extra></extra>',
                                                customdata=np.sqrt(U**2 + V**2 + W**2).flatten()))



        kwargs['axis_equal'] = False
        kwargs['zlim'] = [-spacing,spacing]


    ###############################################################
    ########### DIMENSION: 3 - VECTOR-DIMENSION: 1 ################
    ###############################################################

    elif self.dim == 3 and vector_field.shape == (1,self.xRes,self.yRes,self.zRes):

        ax = tool_plotly_define_3D_plot_ax(fig, ax)

        X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')              

        # Define the vector field
        U = vector_field[0]
        V = np.zeros(U.shape)
        W = np.zeros(U.shape)

        # Add spacing
        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2))

        ax['vmin'] = kwargs.get('vmin', 0)
        ax['vmax'] = kwargs.get('vmax', np.max(max_vector))

        U = U / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*self.dx/self.a0)

        # Scaling
        U = vx_scale*U

        if not kwargs['field_is_nan']:
            fig.add_trace(go.Cone(x=X.flatten()/self.a0, 
                                    y=Y.flatten()/self.a0, 
                                    z=Z.flatten()/self.a0, 
                                    u=U.flatten(), 
                                    v=V.flatten(), 
                                    w=W.flatten(), 
                                    colorscale=kwargs['colormap_object'], 
                                    sizemode='scaled', 
                                    sizeref=1, 
                                    showscale=False,
                                    scene=ax['sceneN'],
                                    hovertemplate='<b>x:</b> %{x:.1f} a₀ <br>' +
                                                    '<b>y:</b> %{y:.1f} a₀ <br>' +
                                                    '<b>z:</b> %{z:.1f} a₀ <br>' +
                                                    '<b>ux:</b> %{u:.1f} <br>' +
                                                    '<b>uy:</b> %{v:.1f} <br>' +
                                                    '<b>uz:</b> %{w:.1f} <br>' +
                                                    '<b>|u|:</b> %{customdata:.1f}<extra></extra>',
                                                    customdata=np.sqrt(U**2 + V**2 + W**2).flatten()))

    ###############################################################
    ########### DIMENSION: 3 - VECTOR-DIMENSION: 2 ################
    ###############################################################

    elif self.dim == 3 and vector_field.shape == (2,self.xRes,self.yRes,self.zRes):

        ax = tool_plotly_define_3D_plot_ax(fig, ax)

        X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')

        # Define the vector field
        U = vector_field[0]
        V = vector_field[1]
        W = np.zeros(U.shape)

        # Add spacing
        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        # Normalize the vectors
        max_vector = np.max(np.sqrt(U ** 2 + V ** 2))

        ax['vmin'] = kwargs.get('vmin', 0)
        ax['vmax'] = kwargs.get('vmax', np.max(max_vector))

        U = U / max_vector
        V = V / max_vector

        # Scale factors
        vx_scale = kwargs.get('vx_scale', spacing*self.dx/self.a0)
        vy_scale = kwargs.get('vy_scale', spacing*self.dy/self.a0)

        # Scaling
        U = vx_scale*U
        V = vy_scale*V

        if not kwargs['field_is_nan']:
            fig.add_trace(go.Cone(x=X.flatten()/self.a0, 
                                    y=Y.flatten()/self.a0, 
                                    z=Z.flatten()/self.a0, 
                                    u=U.flatten(), 
                                    v=V.flatten(), 
                                    w=W.flatten(), 
                                    colorscale=kwargs['colormap_object'], 
                                    sizemode='scaled', 
                                    sizeref=1, 
                                    showscale=False,
                                    scene=ax['sceneN'],
                                    hovertemplate='<b>x:</b> %{x:.1f} a₀ <br>' +
                                                    '<b>y:</b> %{y:.1f} a₀ <br>' +
                                                    '<b>z:</b> %{z:.1f} a₀ <br>' +
                                                    '<b>ux:</b> %{u:.1f} <br>' +
                                                    '<b>uy:</b> %{v:.1f} <br>' +
                                                    '<b>uz:</b> %{w:.1f} <br>' +
                                                    '<b>|u|:</b> %{customdata:.1f}<extra></extra>',
                                                    customdata=np.sqrt(U**2 + V**2 + W**2).flatten()))

    ###############################################################
    ########### DIMENSION: 3 - VECTOR-DIMENSION: 3 ################
    ###############################################################

    elif self.dim == 3 and vector_field.shape == (3,self.xRes,self.yRes,self.zRes):

        ax = tool_plotly_define_3D_plot_ax(fig, ax)

        X, Y, Z = np.meshgrid(self.x, self.y, self.z, indexing='ij')

        # Define the vector field
        U = vector_field[0]
        V = vector_field[1]
        W = vector_field[2]

        X,Y,Z,U,V,W = tool_add_spacing_3D(X,Y,Z,U,V,W,spacing)

        max_vector = np.max(np.sqrt(U ** 2 + V ** 2 + W ** 2))
        ax['vmin'] = kwargs.get('vmin', 0)
        ax['vmax'] = kwargs.get('vmax', np.max(max_vector))

        if not kwargs['field_is_nan']:
            fig.add_trace(go.Cone(x=X.flatten()/self.a0, 
                                    y=Y.flatten()/self.a0, 
                                    z=Z.flatten()/self.a0, 
                                    u=U.flatten(), 
                                    v=V.flatten(), 
                                    w=W.flatten(), 
                                    colorscale=kwargs['colormap_object'], 
                                    sizemode='scaled', 
                                    sizeref=1, 
                                    showscale=False,
                                    scene=ax['sceneN'],
                                    hovertemplate='<b>x:</b> %{x:.1f} a₀ <br>' +
                                                    '<b>y:</b> %{y:.1f} a₀ <br>' +
                                                    '<b>z:</b> %{z:.1f} a₀ <br>' +
                                                    '<b>ux:</b> %{u:.1f} <br>' +
                                                    '<b>uy:</b> %{v:.1f} <br>' +
                                                    '<b>uz:</b> %{w:.1f} <br>' +
                                                    '<b>|u|:</b> %{customdata:.1f}<extra></extra>',
                                                    customdata=np.sqrt(U**2 + V**2 + W**2).flatten()))

    ###############################################################
    ###########     NON-VALID DIMENSION            ################
    ###############################################################

    else:
        raise Exception("You have entered an invalid field to the plot_vector_field function.")


    if kwargs['colorbar'] and not(ax['colorbar']) and not(kwargs['field_is_nan']):
        ax['colormap_object'] = kwargs['colormap_object']
        fig.add_trace(tool_plotly_colorbar(ax, type='normal'))
        ax['colorbar'] = True

    kwargs['fig'] = fig
    kwargs['ax'] = ax
    tool_set_plot_axis_properties_plotly(self, **kwargs)
    return fig, ax


from typing import TYPE_CHECKING, Optional, Any, Tuple

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

import numpy as np
import matplotlib.pyplot as plt
import scipy as sp
from skimage.measure import marching_cubes
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.figure
import matplotlib.axes

from comfit.tool import (
    tool_complete_field,
    tool_set_plot_axis_properties_matplotlib,
    tool_colormap,
    tool_matplotlib_define_3D_plot_ax
)

def plot_field_in_plane_matplotlib(
        self: 'BaseSystem',
        field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs: Any
    ) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Plot the field in a plane perpendicular to the given normal vector.

    Uses scipy.interpolate.griddata and plt.plot_trisurf to visualize the field
    values on a plane defined by its normal vector and a point in space.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    field : np.ndarray
        The field to be plotted.
    normal_vector : np.ndarray, optional
        The normal vector of the plane. Defaults to [0,1,0].
    position : np.ndarray, optional
        The position of the plane. Defaults to the middle of the system.
    \*\*kwargs : Any
        Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

    Returns
    -------
    Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]
        The figure and axes containing the plot.

    Raises
    ------
    Exception
        If the field dimension is not 3D.
    """

    # print('\033[93mWarning: The plot_in_plane function is not yet supported with plotly.\033[0m')

    if self.dim != 3:
        raise Exception("The plot in plane function is only defined for 3D fields.")

    field, fig, ax, kwargs = self.plot_prepare(field, field_type = 'real', **kwargs)

    ax = tool_matplotlib_define_3D_plot_ax(fig, ax) 

    # Default values of position and normal vector
    if position is None:
        position = self.rmid

    if normal_vector is None:
        normal_vector=[0,1,0]

    if ax is None:
        ax = fig.add_subplot(111, projection='3d')

    normal_vector = np.array(normal_vector)/np.linalg.norm(normal_vector)
    height_above_plane = (self.x-position[0])*normal_vector[0] + (self.y-position[1])*normal_vector[1] + (self.z-position[2])*normal_vector[2]

    verts, faces, _, _ = marching_cubes(height_above_plane, 0)

    # Calculate the centroids of each triangle
    centroids = np.mean(verts[faces], axis=1)

    # Assuming field is defined on the same grid as height_above_plane
    x, y, z = np.mgrid[0:height_above_plane.shape[0], 0:height_above_plane.shape[1], 0:height_above_plane.shape[2]]

    # Flatten the grid for interpolation
    points = np.c_[x.ravel(), y.ravel(), z.ravel()]
    field_values = field.ravel()

    # Interpolate field at the vertices positions
    field_verts = sp.interpolate.griddata(points, field_values, centroids, method='nearest')

    # Normalize field values for color mapping
    field_normalized = (field_verts - np.min(field_verts)) / (np.max(field_verts) - np.min(field_verts))

    # Map normalized field values to colors
    colors = kwargs['colormap_object'](field_normalized)

    ax.plot_trisurf((self.xmin+verts[:, 0]*self.dx)/self.a0,
                    (self.ymin+verts[:, 1]*self.dy)/self.a0,
                    faces,
                    (self.zmin+verts[:, 2]*self.dz)/self.a0,
                    facecolor=colors, antialiased=False)

    if kwargs['colorbar']:
            sm = plt.cm.ScalarMappable(cmap=kwargs['colormap_object'])
            sm.set_clim(kwargs['vmin'], kwargs['vmax'])
            plt.colorbar(sm, ax=ax, pad=0.2)

    kwargs['ax'] = ax
    ax = tool_set_plot_axis_properties_matplotlib(self, **kwargs)
    return fig, ax


from typing import TYPE_CHECKING, Any, List, Dict, Tuple

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

import numpy as np

import plotly.graph_objects as go
import plotly.figure_factory as ff

from comfit.tool import (
    tool_set_plot_axis_properties_plotly,
    tool_plotly_define_2D_plot_ax,
    tool_plotly_define_3D_plot_ax,
    tool_extract_node_arrays
)

def plot_nodes_plotly(
    self: 'BaseSystem',
    nodes: List[Dict],
    **kwargs: Any
) -> Tuple[go.Figure, Dict]:
    """Plot the nodes in the system.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    nodes : List[Dict]
        A list of dictionaries representing nodes. Each dictionary must contain:
            - 'position': The position of the node in the defect density array
        Optional keys:
            - 'velocity': The velocity of the node
            - 'charge': The charge of the node
            - 'tangent_vector': The tangent of the node (3D only)
            - 'Burgers_vector': The Burgers vector of the node (for dislocations)
    \*\*kwargs : Any
        Keyword arguments for the plot. See https://comfitlib.com/ClassBaseSystem/
        for a full list of keyword arguments.

    Returns
    -------
    Tuple[go.Figure, Dict]
        The figure object and axes dictionary containing the plot configuration.
    """

    # Check if an axis object is provided
    fig = kwargs.get('fig', go.Figure())
    ax = kwargs.get('ax', {'row': 1, 'col': 1, 'nrows': 1, 'ncols': 1})

    # Check if there are nodes to be plotted
    if not nodes:
        return fig, ax

    # Extract node arrays
    node_arrays = tool_extract_node_arrays(self, nodes)

    if self.dim == 2:

        ax = tool_plotly_define_2D_plot_ax(fig, ax)

        if node_arrays['charge_given']:
            if len(node_arrays['x_coordinates_negative']) > 0:
                fig.add_trace(go.Scatter(x=np.array(node_arrays['x_coordinates_negative'])/self.a0, 
                                        y=np.array(node_arrays['y_coordinates_negative'])/self.a0, 
                                        mode='markers', 
                                        marker=dict(symbol='star', color='blue'),
                                        showlegend=False,
                                        xaxis=ax['xN'],
                                        yaxis=ax['yN']))

            if len(node_arrays['x_coordinates_positive']) > 0:
                fig.add_trace(go.Scatter(x=np.array(node_arrays['x_coordinates_positive'])/self.a0, 
                                        y=np.array(node_arrays['y_coordinates_positive'])/self.a0, 
                                        mode='markers', 
                                        marker=dict(symbol='cross', color='red'),
                                        showlegend=False, 
                                        xaxis=ax['xN'],
                                        yaxis=ax['yN']))
        
        else: # If no charge is provided, plot all nodes as positive
            fig.add_trace(go.Scatter(x=np.array(node_arrays['x_coordinates'])/self.a0, 
                                    y=np.array(node_arrays['y_coordinates'])/self.a0, 
                                    mode='markers', 
                                    marker=dict(symbol='circle', color='black'),
                                    showlegend=False,
                                    xaxis=ax['xN'],
                                    yaxis=ax['yN']))
        
        # Velocities
        if node_arrays['velocity_given']:
            
            vx = np.array(node_arrays['velocity_x_coordinates'])
            vy = np.array(node_arrays['velocity_y_coordinates'])

            velocity_scale = 0.1*min(self.size_x, self.size_y)/self.a0

            vnorm = np.sqrt(np.max(vx**2 + vy**2))
                
            vx = velocity_scale*vx/vnorm
            vy = velocity_scale*vy/vnorm

            fig1 = ff.create_quiver(x=np.array(node_arrays['x_coordinates'])/self.a0, 
                                    y=np.array(node_arrays['y_coordinates'])/self.a0, 
                                    u=vx, 
                                    v=vy, 
                                    line=dict(width=1, color='black'),
                                    scale=1,
                                    showlegend=False,
                                    xaxis=ax['xN'],
                                    yaxis=ax['yN'])
                
            fig.add_traces(data=fig1.data)

        # Burgers vectors
        if node_arrays['Burgers_vector_given']:

            bvx = np.array(node_arrays['Burgers_vector_x_coordinates'])
            bvy = np.array(node_arrays['Burgers_vector_y_coordinates'])

            Burgers_vector_scale = 0.1*min(self.size_x, self.size_y)/self.a0

            bnorm = np.sqrt(np.max(bvx**2 + bvy**2))
                
            bvx = Burgers_vector_scale*bvx/bnorm
            bvy = Burgers_vector_scale*bvy/bnorm

            fig1 = ff.create_quiver(x=np.array(node_arrays['x_coordinates'])/self.a0, 
                                    y=np.array(node_arrays['y_coordinates'])/self.a0, 
                                    u=bvx, 
                                    v=bvy, 
                                    line=dict(width=1, color='red'),
                                    scale=1,
                                    showlegend=False,
                                    xaxis=ax['xN'],
                                    yaxis=ax['yN'])
                
            fig.add_traces(data=fig1.data)

    elif self.dim == 3:

        ax = tool_plotly_define_3D_plot_ax(fig, ax)

        # Plotting options
        quiver_scale = 2 # The scale of the quiver arrows

        tx = np.array(node_arrays['tangent_vector_x_coordinates'])
        ty = np.array(node_arrays['tangent_vector_y_coordinates'])
        tz = np.array(node_arrays['tangent_vector_z_coordinates'])

        vx = np.array(node_arrays['velocity_x_coordinates'])
        vy = np.array(node_arrays['velocity_y_coordinates'])
        vz = np.array(node_arrays['velocity_z_coordinates'])

        if not len(vx) == 0:
            v2 =vx**2 + vy**2 + vz**2
            v_norm = np.sqrt(max(v2))
        else:
            v_norm = 1

        if len(node_arrays['x_coordinates']) > 0:
            #ax.scatter(node_arrays['x_coordinates'], node_arrays['y_coordinates'], z_coords, marker='o', color='black')\
            # fig.add_trace(go.Scatter(x=node_arrays['x_coordinates']/self.a0, 
            #                             y=node_arrays['y_coordinates']/self.a0, 
            #                             mode='markers', 
            #                             marker=dict(symbol='circle', color='black')))   #TODO: Check if this works

            fig.add_trace(go.Cone(x=node_arrays['x_coordinates']/self.a0, 
                                    y=node_arrays['y_coordinates']/self.a0, 
                                    z=z_coords/self.a0, 
                                    u=tx, 
                                    v=ty, 
                                    w=tz, 
                                    scene=ax['sceneN'],
                                    colorscale='Blues', 
                                    sizemode='scaled', 
                                    sizeref=0))

            fig.add_trace(go.Cone(x=node_arrays['x_coordinates']/self.a0, 
                                    y=node_arrays['y_coordinates']/self.a0, 
                                    z=z_coords/self.a0, 
                                    u=vx, 
                                    v=vy, 
                                    w=vz, 
                                    scene=ax['sceneN'],
                                    colorscale='Greens', 
                                    sizemode='scaled', 
                                    sizeref=0))
        

    kwargs['fig'] = fig
    kwargs['ax'] = ax
    tool_set_plot_axis_properties_plotly(self, **kwargs)

    return fig, ax


from comfit.phase_field_crystal.phase_field_crystal import PhaseFieldCrystal

import numpy as np
from tqdm import tqdm
from scipy.optimize import fsolve
import scipy as sp
import matplotlib.pyplot as plt
from pprint import pprint

class PhaseFieldCrystal1DPeriodic(PhaseFieldCrystal):
    def __init__(self, nx, **kwargs):
        """
        Initializes a phase field crystal system in 1D with a periodic crystal structure.

        Parameters
        ----------
        nx : int
            The number of unit cells in the x direction.
        \*\*kwargs : dict, optional
            Additional parameters to configure the system.

        Returns
        -------
        PhaseFieldCrystal1DPeriodic
            The system object representing the simulation.
        """

        # Type of the system
        self.type = 'PhaseFieldCrystal1DPeriodic'
        self.dim = 1
        self.nx = nx

        # Default simulation parameters
        self.micro_resolution = kwargs.get('micro_resolution',[5])
        self.psi0 = -0.3
        self.r = -0.3
        self.t = 0
        self.v = 1
        self.dt = 0.1

        # If there are additional arguments provided, set them as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.xRes = nx * self.micro_resolution[0]

        a0 = 2 * np.pi
        self.a = a0 * np.array([[1]])
        self.q = np.array([[1]])

        # Set the number of reciprocal modes
        self.number_of_reciprocal_lattice_modes = 1
        self.number_of_primary_reciprocal_lattice_modes = 1

        # Set the grid
        self.dx = a0 / self.micro_resolution[0]

        self.type_of_evolution = kwargs.get('type_of_evolution', 'conserved')
        self.A = self.calc_proto_amplitudes_conserved()
        self.eta0 = np.array([self.A])

        # Set the elastic constants
        self.el_mu = 3 * self.A ** 2
        self.el_lambda = 3 * self.A ** 2
        self.el_gamma = 0

        # Initialize the BaseSystem
        kwargs['xRes'] = self.xRes
        kwargs['dx'] = self.dx
        kwargs['dt'] = self.dt
        super().__init__(self.dim, **kwargs)

        # Set the a0
        self.a0 = a0
        self.bool_has_defined_length_scale = True

    def calc_proto_amplitudes_conserved(self):
        """
        Calculates the proto-amplitudes for the system.
        
        Returns
        -------
        float
            The proto-amplitudes for the system.
        """

        psi0 = self.psi0
        r = self.r
        t = self.t
        v = self.v

        A = 0
        
        A_tmp = 1/np.sqrt(3*v)*np.sqrt(- r - 2*t*psi0 - 3*v*psi0**2)
        if self.calc_free_energy_from_proto_amplitudes(psi0, A_tmp) < self.calc_free_energy_from_proto_amplitudes(psi0, A):
            A = A_tmp

        return A

    def calc_free_energy_from_proto_amplitudes(self, psi0, A):
        """
        Calculates the free energy of the system from the proto-amplitudes.

        Parameters
        ----------
        psi0 : float
            The average value of psi.
        A : float
            The proto-amplitude.

        Returns
        -------
        float
            The free energy of the system.
        """

        r = self.r
        t = self.t
        v = self.v

        return 2*np.pi*(1/2*psi0**2 + 1/2*r*(psi0**2 + 2*A**2) + 1/3*t*(psi0**3 + 6*psi0*A**2) + 1/4*v*(psi0**4 + 12*psi0**2*A**2 + 6*A**4))

    def calc_L_f(self):
        """
        Calculates the L operator in Fourier space.

        Returns
        -------
        ndarray
            The L operator in Fourier space.
        """
        k2 = self.calc_k2()
        return 1 - k2

    def calc_omega_f(self):
        """
        Calculates the free energy of the system.

        Returns
        -------
        ndarray
            The free energy of the system.
        """
        
        k2 = self.calc_k2()
        return - k2 * (self.r + (1 - k2)**2)


from typing import List, Tuple, Union, Literal

# General packages
import numpy as np
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt

def tool_colormap(
        colormap_string : str, 
        plot_lib : Literal['plotly', 'matplotlib'] = 'plotly'
        ) -> Union[mcolors.LinearSegmentedColormap, List[Tuple[float, str]]]:
    """Get a colormap for a plot.

    Parameters
    ----------
    colormap_string : str
        The name of the colormap. See options at
        https://comfitlib.com/Plotting/
    plot_lib : str, optional
        The plotting library to use, either 'plotly' or 'matplotlib', by default 'plotly'
    
    Returns
    -------
    Union[mcolors.LinearSegmentedColormap, List[Tuple[float, str]]]
        For plotly: List of tuples with (position, rgb color string)
        For matplotlib: A matplotlib colormap object
    """
    
    
    # Custom colormaps
    sunburst_colors = [  (0.24, 0.15, 0.66),
                    (0.20, 0.48, 0.99),
                    (0.07, 0.74, 0.72),
                    (0.79, 0.75, 0.15),
                    (0.98, 0.98, 0.08)]


    if colormap_string == 'angle':
        # Create hues ranging from cyan to red and back to cyan
        # Note: In the HSV color space, cyan is at 0.5 and red is at 0.
        hues = np.linspace(0.5, 0, 128).tolist() + np.linspace(1, 0.5, 128).tolist()

        # Reverse the order of the hues
        hues = hues[::-1]

        # Convert these hues to RGB colors with full saturation and value
        colors = [mcolors.hsv_to_rgb([h, 1, 1]) for h in hues]

        colormap = mcolors.LinearSegmentedColormap.from_list("custom_hsv", colors)

        if plot_lib == 'plotly':
            # # Convert the colormap to Plotly format
            custom_hsv = []
            for i in range(colormap.N):
                rgb = colormap(i)[:3]
                custom_hsv.append([i / (colormap.N - 1), f'rgb({int(rgb[0]*255)}, {int(rgb[1]*255)}, {int(rgb[2]*255)})'])
            return custom_hsv
        else:
            return colormap

    elif colormap_string == 'angle_r':
        # Create hues ranging from cyan to red and back to cyan
        # Note: In the HSV color space, cyan is at 0.5 and red is at 0.
        hues = np.linspace(0.5, 0, 128).tolist() + np.linspace(1, 0.5, 128).tolist()

        # Convert these hues to RGB colors with full saturation and value
        colors = [mcolors.hsv_to_rgb([h, 1, 1]) for h in hues]

        colormap = mcolors.LinearSegmentedColormap.from_list("custom_hsv_r", colors)

        if plot_lib == 'plotly':
            # # Convert the colormap to Plotly format
            custom_hsv = []
            for i in range(colormap.N):
                rgb = colormap(i)[:3]
                custom_hsv.append([i / (colormap.N - 1), f'rgb({int(rgb[0]*255)}, {int(rgb[1]*255)}, {int(rgb[2]*255)})'])
            return custom_hsv
        else:
            return colormap

    elif colormap_string == 'bluewhitered':
        # Create hues ranging from cyan to red and back to cyan
        # Define the blue-white-red colormap
        colors = [(0, 'blue'), (0.5, 'white'), (1, 'red')]

        if plot_lib == 'plotly':
            return colors
        else:
            return mcolors.LinearSegmentedColormap.from_list('bluewhitered', colors)
    
    elif colormap_string == 'bluewhitered_r':
        # Create hues ranging from cyan to red and back to cyan
        # Define the blue-white-red colormap
        colors = [(0, 'red'), (0.5, 'white'), (1, 'blue')]

        if plot_lib == 'plotly':
            return colors
        else:
            return mcolors.LinearSegmentedColormap.from_list('bluewhitered_r', colors)

    elif colormap_string == 'sunburst':

        colors = sunburst_colors
        
        if plot_lib == 'plotly':
            cm_data = [(i/4, f'rgb({int(rgb[0]*255)}, {int(rgb[1]*255)}, {int(rgb[2]*255)})') 
                      for i, rgb in enumerate(colors)]
            return cm_data

        else:
            positions = [0, 0.25, 0.5, 0.75, 1]
            return mcolors.LinearSegmentedColormap.from_list('sunburst', list(zip(positions, colors)))

    elif colormap_string == 'sunburst_r':

        colors = sunburst_colors[::-1]

        if plot_lib == 'plotly':
            cm_data = [(i/4, f'rgb({int(rgb[0]*255)}, {int(rgb[1]*255)}, {int(rgb[2]*255)})') 
                      for i, rgb in enumerate(colors)]
            return cm_data
        elif plot_lib == 'matplotlib':
            positions = [0, 0.25, 0.5, 0.75, 1]
            return mcolors.LinearSegmentedColormap.from_list('sunburst_r', list(zip(positions, colors)))

    elif colormap_string == 'winter':

        if plot_lib == 'matplotlib':
            return plt.get_cmap('winter')
        else:
            # Get the matplotlib colormap
            cmap = plt.get_cmap('winter')
            # Convert to plotly format by sampling N points and converting to RGB strings
            N = 255
            plotly_colors = [(i/(N-1), f'rgb({int(255*r)},{int(255*g)},{int(255*b)})') 
                             for i, (r,g,b) in enumerate([cmap(i)[:3] for i in range(N)])]
            return plotly_colors

    elif colormap_string == 'winter_r':
        
        if plot_lib == 'matplotlib':
            return plt.get_cmap('winter_r')
        else:
            # Get the matplotlib colormap
            cmap = plt.get_cmap('winter_r')
            # Convert to plotly format by sampling N points and converting to RGB strings
            N = 255
            plotly_colors = [(i/(N-1), f'rgb({int(255*r)},{int(255*g)},{int(255*b)})') 
                             for i, (r,g,b) in enumerate([cmap(i)[:3] for i in range(N)])]
            return plotly_colors

    else:
        if plot_lib == 'plotly':
            return colormap_string
        else:
            return plt.get_cmap(colormap_string)


from comfit.phase_field_crystal.phase_field_crystal import PhaseFieldCrystal

import numpy as np
from tqdm import tqdm
from scipy.optimize import fsolve
import scipy as sp
import matplotlib.pyplot as plt
from pprint import pprint

from comfit.tool.tool_print_in_color import tool_print_in_color

class PhaseFieldCrystal3DFaceCenteredCubic(PhaseFieldCrystal):
    def __init__(self, nx, ny, nz, **kwargs):
        """
        Initializes a phase field crystal system in 3D with a face centered cubic crystal structure.

        Parameters
        ----------
        nx : int
            The number of unit cells in the x direction.
        ny : int
            The number of unit cells in the y direction.
        nz : int
            The number of unit cells in the z direction.
        \*\*kwargs : dict
            Additional arguments to set as attributes.

        Returns
        -------
        None
        """

        # Type of the system
        self.type = 'PhaseFieldCrystal3DFaceCenteredCubic'
        self.dim = 3
        self.nx = nx
        self.ny = ny
        self.nz = nz

        # Default simulation parameters
        self.micro_resolution = [11, 11, 11]
        self.psi0 = -0.325
        self.r = -0.3
        self.t = 0
        self.v = 1
        self.dt = 0.1

        # If there are additional arguments provided, set them as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.xRes = nx * self.micro_resolution[0]
        self.yRes = ny * self.micro_resolution[1]
        self.zRes = nz * self.micro_resolution[2]

        a0 = 2 * np.pi * np.sqrt(3)
        self.a = a0 / 2 * np.array([[0, 1, 1],
                                    [1, 0, 1],
                                    [1, 1, 0],
                                    [0, -1, 1],
                                    [-1, 0, 1],
                                    [-1, 1, 0]], dtype=float)

        self.q = np.array([[-1, 1, 1],
                           [1, -1, 1],
                           [1, 1, -1],
                           [1, 1, 1],
                           [2, 0, 0],
                           [0, 2, 0],
                           [0, 0, 2]], dtype=float) / np.sqrt(3)

        # Set the number of reciprocal modes
        self.number_of_reciprocal_lattice_modes = 7
        self.number_of_primary_reciprocal_lattice_modes = 4

        # Set the grid
        self.dx = a0 / self.micro_resolution[0]
        self.dy = a0 / self.micro_resolution[1]
        self.dz = a0 / self.micro_resolution[2]

        self.type_of_evolution = kwargs.get('type_of_evolution', 'conserved')
        self.A_proto, self.B_proto = self.calc_proto_amplitudes_conserved()
    
        # Set the elastic constants
        self.el_lambda_proto = 32/81 * self.A_proto ** 2
        self.el_mu_proto = 32/81 * self.A_proto ** 2
        self.el_gamma_proto = 32/81 * (2*self.B_proto**2 - self.A_proto**2)

        bool_is_for_properties_calculation = kwargs.get('for_properties_calculation', False)
        if not bool_is_for_properties_calculation:
            tool_print_in_color('Initiating a 3D fcc PFC model.', 'green')
            kwargs.pop('for_properties_calculation', None)
            pfc = PhaseFieldCrystal3DFaceCenteredCubic(1,1,1,for_properties_calculation=True, type_of_evolution=self.type_of_evolution, **kwargs)
            final_strain, self.psi0, self.A, self.B, self.el_lambda, self.el_mu, self.el_gamma = pfc.calc_strained_amplitudes()  
        else:
            self.A = self.A_proto
            self.B = self.B_proto
            self.el_lambda = self.el_lambda_proto
            self.el_mu = self.el_mu_proto
            self.el_gamma = self.el_gamma_proto
            
        self.eta0 = np.array([self.A, self.A, self.A, self.A, self.B, self.B, self.B])

        # Initialize the BaseSystem
        kwargs['xRes'] = self.xRes
        kwargs['yRes'] = self.yRes
        kwargs['zRes'] = self.zRes
        kwargs['dx'] = self.dx
        kwargs['dy'] = self.dy
        kwargs['dz'] = self.dz
        kwargs['dt'] = self.dt
        super().__init__(self.dim, **kwargs)

        # Set the a0
        self.a0 = a0
        self.bool_has_defined_length_scale = True

        if not bool_is_for_properties_calculation:
            self.conf_apply_distortion([[final_strain,0,0],[0,final_strain,0],[0,0,final_strain]], update_q_and_a_vectors=True)
            self.a0 = self.a0 * (1+final_strain)

        
    def calc_proto_amplitudes_conserved(self):
        """
        Calculates the proto-amplitudes for the system.

        Returns
        -------
        tuple
            The proto-amplitudes for the system.
        """

        A = 0
        B = 0
        free_energy = self.calc_free_energy_from_proto_amplitudes(self.psi0, A, B)

        for A0 in np.linspace(0.01, 1, 10):
            for B0 in [A0/2]:
                [A_tmp, B_tmp] = fsolve(self.calc_proto_amplitude_equations_conserved, [A0, B0])
                free_energy_tmp = self.calc_free_energy_from_proto_amplitudes(self.psi0, A_tmp, B_tmp)

                if free_energy_tmp <= free_energy:
                    A = A_tmp
                    B = B_tmp
                    free_energy = free_energy_tmp

        return A, B

    def calc_proto_amplitude_equations_conserved(self, vars):
        """
        Calculates the equations for the proto-amplitudes for the system in the case of conserved dynamics.

        Parameters
        ----------
        vars : list
            The proto-amplitudes for the system.
        
        Returns
        -------
        list
            The equations for the proto-amplitudes for the system in the case of conserved dynamics.
        """
        psi0 = self.psi0
        r = self.r
        t = self.t
        v = self.v

        A, B = vars
        eq1 = A*( r + 27*A**2*v + 36*B**2*v + 2*t*psi0 + 3*v*psi0**2 + 6*B*(t + 3*v*psi0))
        eq2 = 15*B**3*v + 4*A**2*(t + 3*v*psi0) + B*( r + 48*A**2*v + 2*t*psi0 + 3*v*psi0**2)

        return [eq1, eq2]

    def calc_free_energy_from_proto_amplitudes(self, psi0, A, B):
        """
        Calculates the free energy of the system from the proto-amplitudes.

        Parameters
        ----------
        psi0 : float
            The average value of psi.
        A : float
            The proto-amplitude.
        B : float
            The proto-amplitude.

        Returns
        -------
        float
            The free energy of the system.
        """

        r = self.r
        t = self.t
        v = self.v
        return 2*np.pi**3/np.sqrt(3)*(1944*A**4*v + 810*B**4*v + psi0**2*(32 + 18*r + 12*t*psi0 + 9*v*psi0**2) + 108*B**2*(r + psi0*(2*t + 3*v*psi0)) + 144*A**2*( r + 36*B**2*v + 6*B*(t + 3*v*psi0) + psi0*(2*t + 3*v*psi0)))

    def calc_L_f(self):
        """
        Calculates the L operator in Fourier space.

        Returns
        -------
        numpy.ndarray
            The L operator in Fourier space.
        """
        k2 = self.calc_k2()
        return (1 - k2)*(4/3-k2)

    def calc_L_sum_f(self):
        """
        Calculates the sum of the L operators in Fourier space. Needed for stress calculation functions.

        Returns
        -------
        numpy.ndarray
            The L operator in Fourier space.
        """
        k2 = self.calc_k2()
        return 7/3 - 2*k2


from .core import BaseSystem

from .quantum_mechanics import QuantumMechanics

from .phase_field_crystal import \
    PhaseFieldCrystal1DPeriodic, \
    PhaseFieldCrystal2DTriangular, PhaseFieldCrystal2DSquare, \
    PhaseFieldCrystal3DBodyCenteredCubic, PhaseFieldCrystal3DFaceCenteredCubic, \
    PhaseFieldCrystal3DSimpleCubic
    
from .nematic_liquid_crystal import \
    NematicLiquidCrystal

from .bose_einstein_condensate import BoseEinsteinCondensate

from .plot import plot_field_plotly, plot_field_matplotlib, \
    plot_complex_field_matplotlib, plot_complex_field_plotly, \
    plot_angle_field_matplotlib, plot_angle_field_plotly, \
    plot_surface_matplotlib, \
    plot_vector_field_matplotlib, plot_vector_field_plotly, \
    plot_field_in_plane_plotly, plot_field_in_plane_matplotlib, \
    plot_complex_field_in_plane_plotly, plot_complex_field_in_plane_matplotlib, \
    plot_nodes_matplotlib, plot_nodes_plotly

from .tool import tool_colormap, \
    tool_make_animation_movie, tool_make_animation_gif, \
    tool_zoom_plot_matplotlib, tool_multinom, \
    tool_set_plot_axis_properties_matplotlib, tool_set_plot_axis_properties_plotly, \
    tool_complete_field, tool_add_spacing_2D, tool_add_spacing_3D, \
    tool_print_in_color, \
    tool_plotly_find_next_xN, tool_plotly_define_2D_plot_ax, \
    tool_plotly_find_next_sceneN, tool_plotly_define_3D_plot_ax, \
    tool_extract_node_arrays
        





from typing import Union, Dict, Any

# General packages
import plotly.graph_objects as go

# Local packages
from .tool_plotly_find_next_sceneN import tool_plotly_find_next_sceneN

def tool_plotly_define_3D_plot_ax(
        fig: go.Figure, 
        ax: Dict[str, Any]
        ) -> Dict[str, Any]:
    """Defines and validates parameters for a 3D plot axis in Plotly.

    This function checks and sets required parameters for 3D plotting, including
    validating the plot dimension and assigning a scene number.

    Parameters
    ----------
    fig : plotly.graph_objects.Figure
        The Plotly figure object to which the 3D plot will be added.
    ax : dict
        Dictionary containing axis parameters. Must include or will set:
            - plot_dimension : int
                Must be 3 for 3D plots
            - sceneN : int
                Scene number for the 3D plot, auto-generated if not provided
                
    Returns
    -------
    dict
        Updated axis parameters dictionary with validated 3D plot settings.

    Raises
    ------
    ValueError
        If plot_dimension is not 3.
    """

    ax['plot_dimension'] = ax.get('plot_dimension', 3)
    if ax['plot_dimension'] != 3:
        raise ValueError(f"Plot dimension must be 3 for 3D plots. Got {ax['plot_dimension']}")

    ax['sceneN'] = ax.get('sceneN', tool_plotly_find_next_sceneN(fig))

    return ax


from typing import Tuple

#General packages
import numpy as np

def tool_add_spacing_3D(
        X : np.ndarray,
        Y : np.ndarray,
        Z : np.ndarray,
        U : np.ndarray,
        V : np.ndarray,
        W : np.ndarray,
        spacing : int
        ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Add spacing to a 3D field.

    When working with plot vectors, it is often useful to reduce the number of 
    vectors to make the plot more readable. This function allows you to do that 
    by adding spacing to the field.

    Parameters
    ----------
    X : np.ndarray
        The x-coordinates.
    Y : np.ndarray
        The y-coordinates.
    Z : np.ndarray
        The z-coordinates.
    U : np.ndarray
        The x-components of the field.
    V : np.ndarray
        The y-components of the field.
    W : np.ndarray
        The z-components of the field.
    spacing : int
        The spacing to be added.
    
    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        The spaced x-, y-, z-coordinates, x-, y-, and z-components of the field.
    """
    X = X[::spacing, ::spacing, ::spacing]
    Y = Y[::spacing, ::spacing, ::spacing]
    Z = Z[::spacing, ::spacing, ::spacing]
    U = U[::spacing, ::spacing, ::spacing]
    V = V[::spacing, ::spacing, ::spacing]
    W = W[::spacing, ::spacing, ::spacing]
    return X,Y,Z,U,V,W  


from typing import Optional

# General packages
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


def tool_zoom_plot_matplotlib(
        zoom_factor: float, 
        ax: Optional[plt.Axes] = None
        ) -> None:
    """Zoom in/out on a matplotlib plot by adjusting axis limits.

    Parameters
    ----------
    zoom_factor : float
        Factor by which to zoom. Values > 1 zoom in, values < 1 zoom out.
    ax : plt.Axes, optional
        The axes to apply zoom to. If None, uses current axes, by default None.

    Returns
    -------
    None
        Modifies the plot in place.
    """
    if ax is None:
        ax = plt.gca()

    if isinstance(ax, Axes3D):
        # Zoom for 3D plot
        x_lim = ax.get_xlim()
        y_lim = ax.get_ylim()
        z_lim = ax.get_zlim()

        x_center = (x_lim[0] + x_lim[1]) / 2
        y_center = (y_lim[0] + y_lim[1]) / 2
        z_center = (z_lim[0] + z_lim[1]) / 2

        ax.set_xlim3d([x_center - (x_center - x_lim[0]) / zoom_factor,
                       x_center + (x_lim[1] - x_center) / zoom_factor])
        ax.set_ylim3d([y_center - (y_center - y_lim[0]) / zoom_factor,
                       y_center + (y_lim[1] - y_center) / zoom_factor])
        ax.set_zlim3d([z_center - (z_center - z_lim[0]) / zoom_factor,
                       z_center + (z_lim[1] - z_center) / zoom_factor])
    else:
        # Zoom for 2D plot
        x_lim = ax.get_xlim()
        y_lim = ax.get_ylim()

        x_center = (x_lim[0] + x_lim[1]) / 2
        y_center = (y_lim[0] + y_lim[1]) / 2

        ax.set_xlim([x_center - (x_center - x_lim[0]) / zoom_factor,
                     x_center + (x_lim[1] - x_center) / zoom_factor])
        ax.set_ylim([y_center - (y_center - y_lim[0]) / zoom_factor,
                     y_center + (y_lim[1] - y_center) / zoom_factor])





from typing import TYPE_CHECKING, Any, Dict, List, Tuple

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

import numpy as np
import matplotlib.pyplot as plt

from comfit.tool import (
    tool_set_plot_axis_properties_matplotlib,
    tool_extract_node_arrays
)

def plot_nodes_matplotlib(
        self: 'BaseSystem',
        nodes: List[Dict],
        **kwargs: Any
        ) -> Tuple[plt.Figure, plt.Axes]:
    """Plot the nodes of the system.

    Create a matplotlib plot of the nodes in the system, including their positions,
    charges, velocities and Burgers vectors if available.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    nodes : List[Dict]
        List of node dictionaries containing position and property information
    \*\*kwargs : Any
        Keyword arguments for customizing the plot. See 
        https://comfitlib.com/ClassBaseSystem/ for full list.

    Returns
    -------
    Tuple[plt.Figure, plt.Axes]
        The figure and axes objects containing the plot
    """

    # Check if an axis object is provided
    fig = kwargs.get('fig', plt.gcf())
    ax = kwargs.get('ax', None)

    # Check if there are nodes to be plotted, if not return the axes
    if not nodes:
        return fig, ax

    node_arrays = tool_extract_node_arrays(self, nodes)
    
    if self.dim == 2:

        if ax == None:
            fig.clf()
            ax = fig.add_subplot(111)

        x_coords = np.array(node_arrays['x_coordinates'])
        y_coords = np.array(node_arrays['y_coordinates'])

        if node_arrays['charge_given']:
            x_coords_positive = np.array(node_arrays['x_coordinates_positive'])
            y_coords_positive = np.array(node_arrays['y_coordinates_positive'])

            ax.scatter(x_coords_positive/self.a0, y_coords_positive/self.a0, marker='+', color='red')

            x_coords_negative = np.array(node_arrays['x_coordinates_negative'])
            y_coords_negative = np.array(node_arrays['y_coordinates_negative'])

            ax.scatter(x_coords_negative/self.a0, y_coords_negative/self.a0, marker='o', color='blue')
            
        else:
            ax.scatter(x_coords/self.a0, y_coords/self.a0, marker='o', color='black')

        if node_arrays['velocity_given']:
            vx_coords = np.array(node_arrays['velocity_x_coordinates'])
            vy_coords = np.array(node_arrays['velocity_y_coordinates'])
            ax.quiver(x_coords/self.a0, y_coords/self.a0, vx_coords, vy_coords, color='black')

        if node_arrays['Burgers_vector_given']:
            Bx_coords = np.array(node_arrays['Burgers_vector_x_coordinates'])
            By_coords = np.array(node_arrays['Burgers_vector_y_coordinates'])
            ax.quiver(x_coords/self.a0, y_coords/self.a0, Bx_coords, By_coords, color='red')

    elif self.dim == 3:
        # Plotting options
        quiver_scale = 2 # The scale of the quiver arrows

        x_coords = node_arrays['x_coordinates']
        y_coords = node_arrays['y_coordinates']
        z_coords = node_arrays['z_coordinates']

        if node_arrays['tangent_vector_given']:
            tx = np.array(node_arrays['tangent_vector_x_coordinates'])
            ty = np.array(node_arrays['tangent_vector_y_coordinates'])
            tz = np.array(node_arrays['tangent_vector_z_coordinates'])
        
        if node_arrays['velocity_given']:
            vx = np.array(node_arrays['velocity_x_coordinates'])
            vy = np.array(node_arrays['velocity_y_coordinates'])
            vz = np.array(node_arrays['velocity_z_coordinates'])

        if node_arrays['Burgers_vector_given']:
            Bx = np.array(node_arrays['Burgers_vector_x_coordinates'])
            By = np.array(node_arrays['Burgers_vector_y_coordinates'])
            Bz = np.array(node_arrays['Burgers_vector_z_coordinates'])

        if not len(Bx) == 0:
            B2 =Bx**2 + By**2 + Bz**2
            B_norm = np.sqrt(max(B2))
        else:
            B_norm = 1

        ax.scatter(x_coords, y_coords, z_coords, marker='o', color='black')
        if tangent_vector_given:
            ax.quiver(x_coords, y_coords, z_coords, quiver_scale*tx, quiver_scale*ty, quiver_scale*tz, color='blue')
        
        if velocity_given:
            ax.quiver(x_coords, y_coords, z_coords, quiver_scale*vx/v_norm, quiver_scale*vy/v_norm, quiver_scale*vz/v_norm, color='green')
        
        if Burgers_vector_given:
            ax.quiver(x_coords, y_coords, z_coords, quiver_scale*Bx/B_norm, quiver_scale*By/B_norm, quiver_scale*Bz/B_norm, color='red')

    kwargs['fig'] = fig
    kwargs['ax'] = ax
    tool_set_plot_axis_properties_matplotlib(self, **kwargs)

    return fig, ax


from setuptools import setup

setup(
    name='comfit',
    version='1.8.5',
    packages=['comfit'],
    package_data={'comfit':['core/*',
                            'tool/*', 
                            'plot/*',
                            'quantum_mechanics/*', 
                            'bose_einstein_condensate/*',
                            'nematic_liquid_crystal/*',
                            'phase_field_crystal/*' ]},
    author='Vidar Skogvoll and Jonas Rønning',
    install_requires=['numpy>=1.21, <2.0.0',
                      'scipy',
                      'matplotlib',
                      'plotly',
                      'kaleido==0.2.1',
                      'scikit-image>=0.18.0,<0.23',
                      'moviepy',
                      'pillow'],
)


from typing import Tuple, Optional

#General packages
import matplotlib.pyplot as plt
try:
    from moviepy import ImageSequenceClip
except ImportError:
    from moviepy.editor import ImageSequenceClip
import os
from datetime import datetime
from PIL import Image


def tool_make_animation_movie(
        counter : int, 
        name : Optional[str] = None, 
        fps : Optional[int] = 24,
        ID : Optional[str] = None,
        delete_png: Optional[bool] = True
        ) -> None:
    """Make an animation from a series of plot img and save it as an MP4 video file.

    Parameters
    ----------
    counter : int
        The number of plot img to include in the animation.
    name : Optional[str], optional
        The filename for the output video, by default None
    fps : Optional[int], optional
        The frames per second for the video, by default 24
    ID : Optional[str], optional
        A unique identifier for the plot image files, by default None
    delete_png : Optional[bool], optional
        Whether to delete the png files after creating the video, by default True

    Returns
    -------
    None
        Saves the video file.
    """

    if name is None:
        name = datetime.now().strftime("%y%m%d_%H%M") + (' - ' + ID if ID else '') + ' - output_video.mp4'
    else:
        name = datetime.now().strftime("%y%m%d_%H%M") + (' - ' + ID if ID else '') + ' - ' + name + '.mp4'

    # List of saved plot filenames
    if ID is None:
        image_files = [f'plot_{counter}.png' for counter in range(counter+1)]
    else:
        image_files = [f'plot_{counter}_{ID}.png' for counter in range(counter+1)]

    # Create the video clip from the image files
    video_clip = ImageSequenceClip(image_files, fps=fps)

    # Save the video
    video_clip.write_videofile(name)

    # Delete the png files
    if delete_png:
        for file in image_files:
            os.remove(file)


def tool_make_animation_gif(
        counter: int,
        name: Optional[str] = None,
        fps: Optional[int] = 24,
        ID: Optional[str] = None,
        delete_png: Optional[bool] = True
        ) -> None:
    """Make an animation from a series of plot img and save it as a GIF file.

    Parameters
    ----------
    counter : int
        The number of plot img to include in the animation.
    name : Optional[str], optional
        The filename for the output GIF, by default None
    fps : Optional[int], optional
        The frames per second for the GIF, by default 24
    ID : Optional[str], optional
        A unique identifier for the plot image files, by default None
    delete_png : Optional[bool], optional
        Whether to delete the png files after creating the GIF, by default True

    Returns
    -------
    None
        Saves the GIF file.
    """

    if name is None:
        name = datetime.now().strftime("%y%m%d_%H%M%S") + (' - ' + ID if ID else '') + ' - output_animation.gif'
    else:
        name = datetime.now().strftime("%y%m%d_%H%M%S") + (' - ' + ID if ID else '') + ' - ' + name + '.gif'

    if ID is None:
        image_files = [f'plot_{counter}.png' for counter in range(counter+1)]
    else:
        image_files = [f'plot_{counter}_{ID}.png' for counter in range(counter+1)]
        
    images = [Image.open(image_file) for image_file in image_files]

    # Find the smallest image dimensions
    min_width = min(image.width for image in images)
    min_height = min(image.height for image in images)

    # Crop images if necessary and issue a warning
    cropped_images = []
    for image in images:
        if image.width != min_width or image.height != min_height:
            print(f"Warning: Image sizes inconsistent. Cropping image {image.filename} to {min_width}x{min_height}. Consider reviewing how plots are saved mid loop. ")
            image = image.crop((0, 0, min_width, min_height))
        cropped_images.append(image)

    # Save the images as a GIF
    cropped_images[0].save(name, save_all=True, append_images=cropped_images[1:], duration=1000/fps, loop=0)

    # Delete the png files
    if delete_png:
        for file in image_files:
            os.remove(file)




# Typing imports
from typing import TYPE_CHECKING, Optional, Any, Tuple
if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

# General packages
import numpy as np
import matplotlib.pyplot as plt
import scipy as sp
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from skimage.measure import marching_cubes

# Comfit packages
from comfit.tool import tool_complete_field, \
                        tool_set_plot_axis_properties_matplotlib, \
                        tool_colormap, \
                        tool_matplotlib_define_2D_plot_ax, \
                        tool_matplotlib_define_3D_plot_ax

def plot_complex_field_in_plane_matplotlib(
        self: 'BaseSystem',
        complex_field: np.ndarray,
        normal_vector: Optional[np.ndarray] = None,
        position: Optional[np.ndarray] = None,
        **kwargs: Any
        ) -> Tuple[Figure, Axes]:
    """Plot the complex field in a plane perpendicular to the given normal vector.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    complex_field : np.ndarray
        The complex field to be plotted.
    normal_vector : np.ndarray, optional
        The normal vector of the plane. Defaults to [0,1,0].
    position : np.ndarray, optional
        The position of the plane. Defaults to the middle of the system.
    \*\*kwargs : Any
        Keyword arguments for the plot, see https://comfitlib.com/Plotting/.

    Returns
    -------
    tuple[Figure, Axes]
        The figure and axes containing the plot.

    Raises
    ------
    Exception
        If the field dimension is not 3D.
    """


    if self.dim != 3:
        raise Exception("The plot in plane function is only defined for 3D fields.")

    complex_field, fig, ax, kwargs = self.plot_prepare(complex_field, field_type = 'complex', **kwargs)

    ax = tool_matplotlib_define_3D_plot_ax(fig, ax)

    kwargs['plot_is_3D'] = True

    # Default values of position and normal vector
    if position is None:
        position = self.rmid

    if normal_vector is None:
        normal_vector = [0,1,0]

    # Calculate the magnitude and phase of the complex field
    rho = np.abs(complex_field)
    theta = np.angle(complex_field)

    normal_vector = np.array(normal_vector)/np.linalg.norm(normal_vector)
    height_above_plane = (self.x-position[0])*normal_vector[0] + (self.y-position[1])*normal_vector[1] + (self.z-position[2])*normal_vector[2]

    verts, faces, _, _ = marching_cubes(height_above_plane, 0)

    # Calculate the centroids of each triangle
    centroids = np.mean(verts[faces], axis=1)

    # Assuming field is defined on the same grid as height_above_plane
    x, y, z = np.mgrid[0:height_above_plane.shape[0], 0:height_above_plane.shape[1], 0:height_above_plane.shape[2]]

    # Flatten the grid for interpolation
    points = np.c_[x.ravel(), y.ravel(), z.ravel()]
    theta_values = theta.ravel()

    # Interpolate field at the vertices positions
    theta_verts = sp.interpolate.griddata(points, theta_values, centroids, method='nearest')
    rho_verts = sp.interpolate.griddata(points, rho.ravel(), centroids, method='nearest')

    # Normalize field values for color mapping
    theta_normalized = (theta_verts+np.pi) / (2*np.pi)

    # Map normalized field values to colors
    colormap = kwargs['colormap_object']
    colors = colormap(theta_normalized)

    # Blend the colors with white according to rho (normalized)
    colors[:,3] = (rho_verts/np.max(rho_verts)).ravel()

    ax.plot_trisurf((self.xmin+verts[:, 0]*self.dx)/self.a0,
                    (self.ymin+verts[:, 1]*self.dy)/self.a0,
                    faces,
                    (self.zmin+verts[:, 2]*self.dz)/self.a0,
                    facecolor=colors, antialiased=True)


    # Create a colorbar
    if kwargs['colorbar']:
        padding=0.2
        mappable = plt.cm.ScalarMappable(cmap=kwargs['colormap_object'])
        mappable.set_array([])
        mappable.set_clim(-np.pi, np.pi)
        cbar = plt.colorbar(mappable, ax=ax, pad=padding)

        cticks = kwargs.get('cticks', [-np.pi, -2*np.pi/3, -np.pi/3, 0, np.pi/3, 2*np.pi/3, np.pi])
        cbar.set_ticks(cticks)

        cticklabelse = kwargs.get('cticklabels', [r'$-\pi$', r'$-2\pi/3$', r'$-\pi/3$', r'$0$', r'$\pi/3$', r'$2\pi/3$', r'$\pi$'])
        cbar.set_ticklabels(cticklabelse)

    kwargs['grid'] = kwargs.get('grid', True)
    kwargs['ax'] = ax
    tool_set_plot_axis_properties_matplotlib(self, **kwargs)

    return fig, ax


from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

# Standard library imports
from typing import Union

# Third-party library imports
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from mpl_toolkits.mplot3d import Axes3D
import plotly.graph_objects as go

# Local application imports
from comfit.tool import (
    tool_complete_field,
    tool_set_plot_axis_properties_matplotlib,
    tool_set_plot_axis_properties_plotly,
    tool_colormap,
    tool_matplotlib_define_2D_plot_ax,
    tool_matplotlib_define_3D_plot_ax
)

from .plot_surface_matplotlib import plot_surface_matplotlib

def plot_field_matplotlib(
        self: 'BaseSystem',
        field: np.ndarray,
        **kwargs: Any
        ) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Plot the given (real) field.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    field : np.ndarray
        The field to be plotted.
    \*\*kwargs : Any
        Keyword arguments for the plot.
        See https://comfitlib.com/ClassBaseSystem/ 
        for a full list of keyword arguments.

    Returns
    -------
    tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]
        - The figure containing the plot
        - The axes containing the plot
    """

    field, fig, ax, kwargs = self.plot_prepare(field, field_type = 'real', **kwargs)

    ###############################################################
    ###################### DIMENSION: 1 ###########################
    ###############################################################

    if self.dim == 1:

        ax = tool_matplotlib_define_2D_plot_ax(fig, ax)

        ax.plot(self.x/self.a0, field)

    ###############################################################
    ###################### DIMENSION: 2 ###########################
    ###############################################################

    if self.dim == 2:
        
        # Keyword arguments particular to the 2D case
        kwargs['grid'] = kwargs.get('grid', False)

        ax = tool_matplotlib_define_2D_plot_ax(fig, ax)
            
        # Set the colormap
        colormap_string = kwargs.get('colormap', 'viridis')
        colormap = tool_colormap(colormap_string, plot_lib='matplotlib')
        
        # Value limits symmetric
        vlim_symmetric = kwargs.get('vlim_symmetric', False)

        X = kwargs.get('X', None)
        Y = kwargs.get('Y', None)

        if X is None or Y is None:
            X, Y = np.meshgrid(self.x, self.y, indexing='ij')

        pcm = ax.pcolormesh(X / self.a0, Y / self.a0, field, shading='gouraud', cmap=colormap)

        xlim = [self.xmin, self.xmax-self.dx]
        ylim = [self.ymin, self.ymax-self.dy]

        limits_provided = False
        if 'xlim' in kwargs:
            xlim = kwargs['xlim']
            limits_provided = True
        else:
            if 'xmin' in kwargs:
                xlim[0] = kwargs['xmin']
                limits_provided = True
            
            if 'xmax' in kwargs:
                xlim[1] = kwargs['xmax']
                limits_provided = True

        if 'ylim' in kwargs:
            ylim = kwargs['ylim']
            limits_provided = True
        else:
            if 'ymin' in kwargs:
                ylim[0] = kwargs['ymin']
                limits_provided = True
                
            if 'ymax' in kwargs:
                ylim[1] = kwargs['ymax']
                limits_provided = True

        # If explicit limits are provided, use them to change the vlim ranges
        if limits_provided:
            region_to_plot = np.zeros(self.dims).astype(bool)
            region_to_plot[(xlim[0] <= X)*(X <= xlim[1])*(ylim[0] <= Y)*(Y <= ylim[1])] = True
            vlim = [np.min(field[region_to_plot]), np.max(field[region_to_plot])]

        else:
            vlim = [np.min(field), np.max(field)]
        
        # Set the value limits
        if 'vlim' in kwargs:
            vlim = kwargs['vlim']
        else:
            if 'vmin' in kwargs:
                vlim[0] = kwargs['vmin']
            if 'vmax' in kwargs:
                vlim[1] = kwargs['vmax']

        if vlim[1] - vlim[0] < 1e-10:
            vlim = [vlim[0]-0.05, vlim[1]+0.05]

        pcm.set_clim(vmin=vlim[0], vmax=vlim[1])

        if 'vlim_symmetric' in kwargs:
            vlim_symmetric = kwargs['vlim_symmetric']
            if vlim_symmetric:
                cmax = abs(field).max()
                cmin = -cmax
                pcm.set_clim(vmin=cmin, vmax=cmax)

        colorbar = kwargs.get('colorbar', True)

        if colorbar:
            cbar = plt.colorbar(pcm, ax=ax)

    ###############################################################
    ###################### DIMENSION: 3 ###########################
    ###############################################################

    elif self.dim == 3:

        ax = tool_matplotlib_define_3D_plot_ax(fig, ax)
        kwargs['plot_is_3D'] = True
            
        # Keyword arguments particular to the 3D case
        number_of_layers = kwargs.get('number_of_layers', 1)
        alpha = kwargs.get('alpha', 0.5)

        vmin = kwargs['vmin']
        vmax = kwargs['vmax']

        if 'layer_values' in kwargs:
            layer_values = np.concatenate([[-np.inf], kwargs['layer_values'], [np.inf]])
        else: 
            layer_values = np.linspace(vmin, vmax, number_of_layers + 2)

        field_min = np.min(field)
        field_max = np.max(field)

        if field_min < layer_values[1] < field_max:
            plot_surface_matplotlib(self, field=field, 
                                    value=layer_values[1], 
                                    color=kwargs['colormap_object']((layer_values[1]-vmin) / (vmax-vmin)), 
                                    alpha=alpha,
                                    ax=ax)

        for layer_value in layer_values[2:-1]:
            if field_min < layer_value < field_max:
                plot_surface_matplotlib(self, field=field, 
                                        value=layer_value, 
                                        color=kwargs['colormap_object']((layer_value-vmin) / (vmax-vmin)), 
                                        alpha=alpha,
                                        ax=ax)

        if kwargs['colorbar']:
            sm = plt.cm.ScalarMappable(cmap=kwargs['colormap_object'])
            sm.set_clim(kwargs['vmin'], kwargs['vmax'])
            plt.colorbar(sm, ax=ax, pad=0.2)

    kwargs['ax'] = ax
    ax = tool_set_plot_axis_properties_matplotlib(self, **kwargs)
    return fig, ax



# General packages
import numpy as np
from scipy.fftpack import fftn, ifftn
from scipy.special import factorial
from collections import Counter

def tool_multinom(
        *args: int
        ) -> float:
    """Calculate the multinomial coefficient.

    Parameters
    ----------
    \*args : int
        Sequence of integers representing the counts of each category.

    Returns
    -------
    float
        The multinomial coefficient value.
    """
    M = len(args)
    if M < 2:
        return 1
    else:
        indices = np.array(args)
        counts = Counter(indices)
        prod = np.prod([factorial(count) for count in counts.values()])
        return 1 / prod

def levi_civita_symbol(i, j, k):
    """Calculate the Levi-Civita symbol (antisymmetric tensor).

    The Levi-Civita symbol is defined as:
        1 if (i,j,k) is an even permutation of (0,1,2)
        -1 if (i,j,k) is an odd permutation of (0,1,2)
        0 if any indices are repeated or not in {0,1,2}
    
    Parameters
    ----------
    i : int
        First index
    j : int
        Second index  
    k : int
        Third index

    Returns
    -------
    int
        Value of Levi-Civita symbol: 1, -1, or 0

    Examples
    --------
    >>> levi_civita_symbol(0,1,2)
    1
    >>> levi_civita_symbol(1,0,2) 
    -1
    >>> levi_civita_symbol(0,0,1)
    0
    """
    if {i, j, k} != {0, 1, 2}:
        return 0  # Levi-Civita symbol is zero if indices are not 0, 1, 2

    if (i, j, k) in [(0, 1, 2), (1, 2, 0), (2, 0, 1)]:
        return 1  # Even permutation
    elif (i, j, k) in [(2, 1, 0), (0, 2, 1), (1, 0, 2)]:
        return -1  # Odd permutation
    else:
        return 0  # Repeated index

if __name__ == "__main__":
    print(tool_multinom(1,2,2))


# Type checking imports
from typing import TYPE_CHECKING, Optional, Any, Tuple

if TYPE_CHECKING:
    from comfit.core.base_system import BaseSystem

# Standard library imports
import numpy as np
import scipy as sp

# Third-party library imports
from skimage.measure import marching_cubes
import plotly.graph_objects as go

# Local application imports
from comfit.tool import (
    tool_matplotlib_define_3D_plot_ax,
    tool_plotly_define_3D_plot_ax,
    tool_set_plot_axis_properties_matplotlib,
    tool_set_plot_axis_properties_plotly,
    tool_plotly_colorbar
)


def plot_vector_field_in_plane_both_plot_libs(
        self: 'BaseSystem',
        vector_field: np.ndarray,
        position: Optional[np.ndarray] = None,
        normal_vector: Optional[np.ndarray] = None,
        spacing: Optional[int] = None,
        **kwargs: Any
        ) -> Tuple[Any, Any]:
    """Plot a vector field on a plane in 3D space.

    Creates a visualization of a vector field where it intersects with a plane
    defined by a point and a normal vector.

    Parameters
    ----------
    self : BaseSystem
        A BaseSystem (or derived) instance.
    vector_field : np.ndarray
        The vector field to plot. Shape must be (1,xRes,yRes,zRes), 
        (2,xRes,yRes,zRes), or (3,xRes,yRes,zRes).
    position : np.ndarray, optional
        Point through which the plane passes. Defaults to system midpoint.
    normal_vector : np.ndarray, optional
        Normal vector defining the plane orientation. Defaults to [1,1,1].
    spacing : int, optional
        Spacing between vectors in the plot. Defaults to max(xRes//20,1).
    \*\*kwargs : Any
        Additional keyword arguments for plot customization. See https://comfitlib.com/Plotting/.

    Returns
    -------
    Tuple[Any, Any]
        Figure and axes objects from the plotting library.

    Raises
    ------
    Exception
        If the system is not 3D or if the vector field has invalid dimensions.
    """

    plot_lib = kwargs.get('plot_lib', self.plot_lib)

    if self.dim != 3:
        raise Exception("The plot_vector_field_in_plane function is only defined for 3D fields.")

    vector_field, fig, ax, kwargs = self.plot_prepare(vector_field, field_type = 'vector', **kwargs)

    # Default values of position and normal vector
    if position is None:
        position = self.rmid
    
    if normal_vector is None:
        normal_vector = [1,1,1]

    if spacing is None:
        spacing = max(self.xRes//20,1)

    normal_vector = np.array(normal_vector)/np.linalg.norm(normal_vector)
    height_above_plane = (self.x-position[0])*normal_vector[0] + (self.y-position[1])*normal_vector[1] + (self.z-position[2])*normal_vector[2]

    verts, faces, _, _ = marching_cubes(height_above_plane, 0)

    # Assuming field is defined on the same grid as height_above_plane
    x, y, z = np.mgrid[0:height_above_plane.shape[0], 0:height_above_plane.shape[1], 0:height_above_plane.shape[2]]

    # Flatten the grid for interpolation
    points = np.c_[x.ravel(), y.ravel(), z.ravel()]

    # Extract the vector field
    if vector_field.shape == (1,self.xRes,self.yRes,self.zRes):
        n = 1
    elif vector_field.shape == (2,self.xRes,self.yRes,self.zRes):
        n = 2
    elif vector_field.shape == (3,self.xRes,self.yRes,self.zRes):
        n = 3
    else:
        raise Exception("You have entered an invalid field to the plot_vector_field function.")

    U = vector_field[0]
    U_values = U.ravel()
    U_verts = sp.interpolate.griddata(points, U_values, verts, method='nearest')

    if n > 1:
        V = vector_field[1]
        V_values = V.ravel()
        V_verts = sp.interpolate.griddata(points, V_values, verts, method='nearest')
    else:
        V_verts = np.zeros(U_verts.shape)
    
    if n > 2:
        W = vector_field[2]
        W_values = W.ravel()
        W_verts = sp.interpolate.griddata(points, W_values, verts, method='nearest')
    else:
        W_verts = np.zeros(U_verts.shape)

    # Normalize the vectors
    max_vector = np.max(np.sqrt(U_verts ** 2 + V_verts ** 2 + W_verts ** 2))
    U_verts = U_verts / max_vector
    V_verts = V_verts / max_vector
    W_verts = W_verts / max_vector

    # Scale factors
    vx_scale = kwargs.get('vx_scale', spacing*self.dx)
    vy_scale = kwargs.get('vy_scale', spacing*self.dy)
    vz_scale = kwargs.get('vz_scale', spacing*self.dz)

    # Scaling
    U_verts = vx_scale*U_verts
    V_verts = vy_scale*V_verts
    W_verts = vz_scale*W_verts

    x = self.xmin+verts[:, 0]*self.dx
    y = self.ymin+verts[:, 1]*self.dy
    z = self.zmin+verts[:, 2]*self.dz

    # Adjust positions based on the coordinate system
    x = self.xmin + x * self.dx
    y = self.ymin + y * self.dy
    z = self.zmin + z * self.dz

    # Add spacing
    x = x[::spacing]
    y = y[::spacing]
    z = z[::spacing]
    U_verts = U_verts[::spacing]
    V_verts = V_verts[::spacing]
    W_verts = W_verts[::spacing]

    if plot_lib == "plotly":

        ax = tool_plotly_define_3D_plot_ax(fig, ax) #Defines sceneN, plot_dimension

        ax['vmin'] = kwargs.get('vmin', 0)
        ax['vmax'] = kwargs.get('vmax', np.max(max_vector))

        if not kwargs['field_is_nan']:
            fig.add_trace(go.Cone(  x=x.flatten()/self.a0, 
                                    y=x.flatten()/self.a0, 
                                    z=y.flatten()/self.a0, 
                                    u=U_verts.flatten(), 
                                    v=V_verts.flatten(), 
                                    w=W_verts.flatten(), 
                                    colorscale=kwargs['colormap_object'], 
                                    sizemode='scaled', 
                                    sizeref=1, 
                                    showscale=False, 
                                    scene = ax['sceneN'],
                                    hovertemplate='<b>x:</b> %{x:.1f} a₀ <br>' +
                                                '<b>ux:</b> %{u:.1f} <br>' +
                                                '<b>uy:</b> %{v:.1f} <br>' +
                                                '<b>uz:</b> %{w:.1f} <br>' +
                                                '<b>|u|:</b> %{customdata:.1f}<extra></extra>',
                                                customdata=np.sqrt(U_verts**2 + V_verts**2 + W_verts**2).flatten()))

        if kwargs['colorbar'] and not(ax['colorbar']) and not(kwargs['field_is_nan']):
            ax['colormap_object'] = kwargs['colormap_object']
            fig.add_trace(tool_plotly_colorbar(ax, type='normal'))
            ax['colorbar'] = True

        kwargs['fig'] = fig
        kwargs['ax'] = ax
        tool_set_plot_axis_properties_plotly(self, **kwargs)

    elif plot_lib == "matplotlib":
        
        ax = tool_matplotlib_define_3D_plot_ax(fig, ax)
        kwargs['plot_is_3D'] = True

        ax.quiver(x, y, z, U_verts, V_verts, W_verts, color='blue')

        kwargs['fig'] = fig
        kwargs['ax'] = ax
        tool_set_plot_axis_properties_matplotlib(self, **kwargs)
    
    return fig, ax


from typing import Optional, Union, Literal, TYPE_CHECKING, List, Dict, Any

# General packages
import numpy as np
import matplotlib.pyplot as plt
import scipy as sp

# Local packages
from comfit.core import BaseSystem
from comfit.tool import tool_colormap

class QuantumMechanics(BaseSystem):
    def __init__(self : 'QuantumMechanics', 
            dimension: int, 
            **kwargs: Dict[str, Any]
            ) -> None:
        """Initializes a quamtum mechanics system evolving according to the Schrödinger equation

        Parameters
        ----------
        dimension : int
            The dimension of the system.
        \*\*kwargs : dict, optional
            Optional keyword arguments to set additional parameters. 
            See https://comfitlib.com/ClassBaseSystem/.

        Examples
        --------
        >>> qm = cf.QuantumMechanics(3,xRes=101,yRes=101,zRes=101)
        Creates a BoseEinsteinCondensate system with 3 dimensions and a spatial resolution of 101.
        """

        # First initialize the BaseSystem class
        super().__init__(dimension, **kwargs)

        # Type of the system
        self.psi = None
        self.psi_f = None
        self.type = 'QuantumMechanics1' #Quantum mechanical system with one particle.

        # Default simulation parameters
        self.V_ext = 0  # External potential

        # If there are additional arguments provided, set them as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(
        self: 'QuantumMechanics',
        ) -> str:
        """Returns a string representation of the QuantumMechanics system.

        This method is called when str() is invoked on an instance of the class.

        Returns
        -------
        str: String representation 'QuantumMechanics' of the system
        """
        return 'QuantumMechanics'

    def conf_initial_condition_Gaussian(
            self : 'QuantumMechanics',
            position: Optional[Union[List[float], np.ndarray]] = None,
            width: Optional[float] = None,
            initial_velocity: Optional[Union[float, list[float]]] = None
            ) -> None:
        """Set the initial condition to a Gaussian wavepacket.

        Parameters
        ----------
        position : array_like, optional
            The position of the wavepacket.
        width : float, optional
            The width of the wavepacket.
        initial_velocity : float or array_like, optional
            The initial velocity of the wavepacket.

        Returns
        -------
        None
            Configures self.psi and self.psi_f.
        """

        if position is None:
            position = self.rmid
        if width is None:
            width = self.size_x/10

        self.psi = np.sqrt(self.calc_Gaussian(position=position,width=width))
        
        if initial_velocity is not None:
            if self.dim == 1:
                v0 = initial_velocity
                self.psi = self.psi * np.exp(1j * v0 * self.x)

            elif self.dim == 2:
                v0 = initial_velocity
                self.psi = self.psi * np.exp(1j * (v0[0] * self.x + v0[1]*self.y))

            elif self.dim == 3:
                v0 = initial_velocity
                self.psi = self.psi * np.exp(1j * (v0[0] * self.x + v0[1]*self.y + v0[2]*self.z))
        
        self.psi_f = sp.fft.fftn(self.psi)


    def conf_harmonic_potential(
            self : 'QuantumMechanics', 
            trapping_strength: Optional[float] = None
            ) -> None:
        """Set the external potential to a harmonic trap with R_tf being the thomas fermi radius
        
        Parameters
        ----------
        trapping_strength : float, optional
            The strength of the trapping potential.

        Returns
        -------
        None       
            Configures self.V_ext.
        """

        if trapping_strength is None:
            trapping_strength = 1/(0.25*self.size_x)**2

        if self.dim == 1:
            return  trapping_strength*(self.x - self.xmid )**2

        if self.dim == 2:
            return trapping_strength*(((self.x-self.xmid)**2) +((self.y-self.ymid)**2) )
        if self.dim == 3:
            return trapping_strength * (((self.x - self.xmid) ** 2)
                                           + ((self.y - self.ymid) ** 2)
                                           +((self.z - self.zmid) ** 2) )

    def conf_hydrogen_state(
            self: 'QuantumMechanics', 
            n: int, 
            l: int, 
            m: int
            ) -> None:
        """Set the initial condition to a hydrogen state with quantum numbers n,l,m

        Parameters
        ----------
        n : int 
            principal quantum number
        l : int
            azimuthal quantum number
        m : int 
            magnetic quantum number

        Returns
        -------
        None
            Configures self.psi and self.psi_f.
        """

        self.psi = self.calc_hydrogen_state(n,l,m)
        self.psi_f = sp.fft.fftn(self.psi)

    def conf_wavefunction(
            self: 'QuantumMechanics', 
            psi: np.ndarray
            ) -> None:
        """Set the initial condition to a given wavefunction.

        Parameters
        ----------
        psi : np.ndarray
            The wavefunction.

        Returns
        -------
        None
            Configures self.psi and self.psi_f.
        """
        self.psi = psi
        self.psi_f = sp.fft.fftn(self.psi)   

    def calc_hydrogen_state(
            self: 'QuantumMechanics', 
            n: int, 
            l: int, 
            m: int
            ) -> np.ndarray:
        """Calculate the hydrogen state with quantum numbers n,l,m

        Parameters
        ----------
        n : int
            principal quantum number
        l : int
            azimuthal quantum number
        m : int
            magnetic quantum number

        Returns
        -------
        np.ndarray
            Wavefunction of the hydrogen state. (np.ndarray)
        """

        if not self.dim == 3:
            raise(Exception('The hydrogen state is only defined in 3D'))

        r = np.sqrt(self.x**2 + self.y**2 + self.z**2)
        theta = np.arccos(self.z/r)
        theta[np.isnan(theta)] = 0
        phi = np.arctan2(self.y,self.x)

        R = (2*r/n)**l * np.exp(-r/n) *  sp.special.genlaguerre(n-l-1,2*l+1)(2*r/n)
        Y = sp.special.sph_harm(m,l,phi,theta)
        
        return R*Y
    

    ## Calculation functions
    def calc_nonlinear_evolution_function_f(
            self: 'QuantumMechanics', 
            psi: np.ndarray, 
            t: float
            ) -> np.ndarray:
        """Calculate the nonlinear evolution function f of the Schrödinger equation

        Parameters
        ----------
        psi : np.ndarray 
            The wavefunction.
        t : float
            Time
        
        Returns
        -------
        np.ndarray
            The nonlinear evolution function f.
        """

        if callable(self.V_ext):
            potential = self.V_ext(t)
        else:
            potential = self.V_ext
        return sp.fft.fftn((1j) * (-potential) * psi)


    ## Time evolution functions
    def evolve_schrodinger(
            self: 'QuantumMechanics', 
            number_of_steps: int, 
            method: Literal['ETD2RK', 'ETD4RK'] = 'ETD4RK'
            ) -> None:
        """Evolve the system according to the Schrödinger equation

        Parameters
        ----------
        number_of_steps : int
            Number of time steps to evolve the system.
        method : str
            The method to use for the time evolution. 
        
        Returns
        -------
        None
            Updates self.psi and self.psi_f.
        """

        omega_f = -1j / 2 * self.calc_k2()
        if method == 'ETD2RK':
            integrating_factors_f = self.calc_evolution_integrating_factors_ETD2RK(omega_f)
            solver = self.evolve_ETD2RK_loop
        elif method == 'ETD4RK':
            integrating_factors_f = self.calc_evolution_integrating_factors_ETD4RK(omega_f)
            solver = self.evolve_ETD4RK_loop
        else:
            raise(Exception('This method is not implemented'))
        for n in range(number_of_steps):
            self.psi, self.psi_f = solver(integrating_factors_f, self.calc_nonlinear_evolution_function_f,
                                                           self.psi, self.psi_f)